# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# alias because we already had an option named argparse
import argparse as argparse_mod
import collections
import copy
import errno
import json
import os
import re
import sys
import typing as ty
import warnings

from keystoneauth1 import adapter
from keystoneauth1 import loading
from keystoneauth1 import session
import platformdirs
import yaml

from openstack import _log
from openstack.config import _util
from openstack.config import cloud_region
from openstack.config import defaults
from openstack.config import vendors
from openstack import exceptions
from openstack import warnings as os_warnings

if ty.TYPE_CHECKING:
    from keystoneauth1.loading._plugins.identity import v3 as v3_loaders
    from keystoneauth1.loading import opts

PLATFORMDIRS = platformdirs.PlatformDirs(
    'openstack', 'OpenStack', multipath=True
)
CONFIG_HOME = PLATFORMDIRS.user_config_dir
CACHE_PATH = PLATFORMDIRS.user_cache_dir

# snaps do set $HOME to something like
# /home/$USER/snap/openstackclients/$SNAP_VERSION
# the real home (usually /home/$USERNAME) is stored in $SNAP_REAL_HOME
# see https://snapcraft.io/docs/environment-variables
SNAP_REAL_HOME = os.getenv('SNAP_REAL_HOME')
if SNAP_REAL_HOME:
    UNIX_CONFIG_HOME = os.path.join(
        os.path.join(SNAP_REAL_HOME, '.config'), 'openstack'
    )
else:
    UNIX_CONFIG_HOME = os.path.join(
        os.path.expanduser(os.path.join('~', '.config')), 'openstack'
    )
UNIX_SITE_CONFIG_HOME = '/etc/openstack'

SITE_CONFIG_HOME = PLATFORMDIRS.site_config_dir

CONFIG_SEARCH_PATH = [
    os.getcwd(),
    CONFIG_HOME,
    UNIX_CONFIG_HOME,
    SITE_CONFIG_HOME,
    UNIX_SITE_CONFIG_HOME,
]
YAML_SUFFIXES = ('.yaml', '.yml')
JSON_SUFFIXES = ('.json',)
CONFIG_FILES = [
    os.path.join(d, 'clouds' + s)
    for d in CONFIG_SEARCH_PATH
    for s in YAML_SUFFIXES + JSON_SUFFIXES
]
SECURE_FILES = [
    os.path.join(d, 'secure' + s)
    for d in CONFIG_SEARCH_PATH
    for s in YAML_SUFFIXES + JSON_SUFFIXES
]
VENDOR_FILES = [
    os.path.join(d, 'clouds-public' + s)
    for d in CONFIG_SEARCH_PATH
    for s in YAML_SUFFIXES + JSON_SUFFIXES
]

BOOL_KEYS = ('insecure', 'cache')
CSV_KEYS = ('auth_methods',)

FORMAT_EXCLUSIONS = frozenset(['password'])


def get_boolean(value: ty.Any) -> bool:
    if value is None:
        return False
    if type(value) is bool:
        return value
    if value.lower() == 'true':
        return True
    return False


def _auth_update(
    old_dict: dict[str, ty.Any], new_dict_source: dict[str, ty.Any]
) -> dict[str, ty.Any]:
    """Like dict.update, except handling the nested dict called auth."""
    new_dict = copy.deepcopy(new_dict_source)
    for k, v in new_dict.items():
        if k == 'auth':
            if k in old_dict:
                old_dict[k].update(v)
            else:
                old_dict[k] = v.copy()
        else:
            old_dict[k] = v
    return old_dict


def _fix_argv(argv: list[str]) -> None:
    # Transform any _ characters in arg names to - so that we don't
    # have to throw billions of compat argparse arguments around all
    # over the place.
    processed = collections.defaultdict(set)
    for index in range(0, len(argv)):
        # If the value starts with '--' and has '-' or '_' in it, then
        # it's worth looking at it
        if re.match('^--.*(_|-)+.*', argv[index]):
            split_args = argv[index].split('=')
            orig = split_args[0]
            new = orig.replace('_', '-')
            if orig != new:
                split_args[0] = new
                argv[index] = "=".join(split_args)
            # Save both for later so we can throw an error about dupes
            processed[new].add(orig)
    overlap: list[str] = []
    for new, old in processed.items():
        if len(old) > 1:
            overlap.extend(old)
    if overlap:
        raise exceptions.ConfigException(
            "The following options were given: '{options}' which contain "
            "duplicates except that one has _ and one has -. There is "
            "no sane way for us to know what you're doing. Remove the "
            "duplicate option and try again".format(options=','.join(overlap))
        )


class OpenStackConfig:
    # These two attribute are to allow os-client-config to plumb in its
    # local versions for backwards compat.
    # They should not be used by anyone else.
    _cloud_region_class = cloud_region.CloudRegion
    _defaults_module = defaults

    #: config_filename is the filename that configuration was loaded from, if
    #: any.
    config_filename: str | None
    #: secure_config_filename is the filename that secure configuration was
    #: loaded from, if any.
    secure_config_filename: str | None
    #: cloud_config contains the combined loaded configuration.
    cloud_config: dict[str, ty.Any]

    def __init__(
        self,
        config_files: list[str] | None = None,
        vendor_files: list[str] | None = None,
        override_defaults: dict[str, ty.Any] | None = None,
        force_ipv4: bool | None = None,
        envvar_prefix: str | None = None,
        secure_files: list[str] | None = None,
        pw_func: cloud_region._PasswordCallback | None = None,
        session_constructor: type[session.Session] | None = None,
        app_name: str | None = None,
        app_version: str | None = None,
        load_yaml_config: bool = True,
        load_envvars: bool = True,
        statsd_host: str | None = None,
        statsd_port: str | None = None,
        statsd_prefix: str | None = None,
        influxdb_config: dict[str, ty.Any] | None = None,
    ):
        self.log = _log.setup_logging('openstack.config')
        self._session_constructor = session_constructor
        self._app_name = app_name
        self._app_version = app_version
        self._load_envvars = load_envvars

        if load_yaml_config:
            # "if config_files" is not sufficient to process empty list
            if config_files is not None:
                self._config_files = config_files
            else:
                self._config_files = CONFIG_FILES
            if secure_files is not None:
                self._secure_files = secure_files
            else:
                self._secure_files = SECURE_FILES
            if vendor_files is not None:
                self._vendor_files = vendor_files
            else:
                self._vendor_files = VENDOR_FILES
        else:
            self._config_files = []
            self._secure_files = []
            self._vendor_files = []

        config_file_override = self._get_envvar('OS_CLIENT_CONFIG_FILE')
        if config_file_override:
            self._config_files.insert(0, config_file_override)

        secure_file_override = self._get_envvar('OS_CLIENT_SECURE_FILE')
        if secure_file_override:
            self._secure_files.insert(0, secure_file_override)

        self.defaults = self._defaults_module.get_defaults()
        if override_defaults:
            self.defaults.update(override_defaults)

        # First, use a config file if it exists where expected
        config_filename, cloud_config = self._load_config_file()
        if config_filename and cloud_config:
            self._validate_config_file(config_filename, cloud_config)

        secure_config_filename, secure_config = self._load_secure_file()
        if secure_config_filename and secure_config:
            self._validate_config_file(secure_config_filename, secure_config)
            cloud_config = _util.merge_clouds(
                cloud_config or {}, secure_config
            )

        self.config_filename = config_filename
        self.secure_config_filename = secure_config_filename
        if not cloud_config:
            self.cloud_config = {'clouds': {}}
        else:
            self.cloud_config = cloud_config
            if 'clouds' not in self.cloud_config:
                self.cloud_config['clouds'] = {}

        # Save the other config
        self.extra_config = copy.deepcopy(self.cloud_config)
        self.extra_config.pop('clouds', None)

        # Grab ipv6 preference settings from env
        client_config = self.cloud_config.get('client', {})

        if force_ipv4 is not None:
            # If it's passed in to the constructor, honor it.
            self.force_ipv4 = force_ipv4
        else:
            # Get the backwards compat value
            prefer_ipv6 = get_boolean(
                self._get_envvar(
                    'OS_PREFER_IPV6',
                    client_config.get(
                        'prefer_ipv6', client_config.get('prefer-ipv6', True)
                    ),
                )
            )
            force_ipv4 = get_boolean(
                self._get_envvar(
                    'OS_FORCE_IPV4',
                    client_config.get(
                        'force_ipv4', client_config.get('broken-ipv6', False)
                    ),
                )
            )

            self.force_ipv4 = force_ipv4
            if not prefer_ipv6:
                # this will only be false if someone set it explicitly
                # honor their wishes
                self.force_ipv4 = True

        # Next, process environment variables and add them to the mix
        self.envvar_key = self._get_envvar('OS_CLOUD_NAME', 'envvars')
        if self.envvar_key in self.cloud_config['clouds']:
            raise exceptions.ConfigException(
                f'{self.config_filename!r} defines a cloud named '
                f'{self.envvar_key!r}, but OS_CLOUD_NAME is also set to '
                f'{self.envvar_key!r}. '
                f'Please rename either your environment-based cloud, '
                f'or one of your file-based clouds.'
            )

        self.default_cloud = self._get_envvar('OS_CLOUD')

        if load_envvars:
            envvars = self._get_os_environ(envvar_prefix=envvar_prefix)
            if envvars:
                self.cloud_config['clouds'][self.envvar_key] = envvars
                if not self.default_cloud:
                    self.default_cloud = self.envvar_key

        if not self.default_cloud and self.cloud_config['clouds']:
            if len(self.cloud_config['clouds'].keys()) == 1:
                # If there is only one cloud just use it. This matches envvars
                # behavior and allows for much less typing.
                # TODO(mordred) allow someone to mark a cloud as "default" in
                # clouds.yaml.
                # The next/iter thing is for python3 compat where dict.keys
                # returns an iterator but in python2 it's a list.
                self.default_cloud = next(
                    iter(self.cloud_config['clouds'].keys())
                )

        # Finally, fall through and make a cloud that starts with defaults
        # because we need somewhere to put arguments, and there are neither
        # config files or env vars
        if not self.cloud_config['clouds']:
            self.cloud_config = dict(clouds=dict(defaults=dict(self.defaults)))
            self.default_cloud = 'defaults'

        self._cache_auth = False
        self._cache_expiration_time = 0
        self._cache_path = CACHE_PATH
        self._cache_class = 'dogpile.cache.null'
        self._cache_arguments: dict[str, ty.Any] = {}
        self._cache_expirations: dict[str, int] = {}
        self._influxdb_config = {}
        if 'cache' in self.cloud_config:
            cache_settings = _util.normalize_keys(self.cloud_config['cache'])

            self._cache_auth = get_boolean(
                cache_settings.get('auth', self._cache_auth)
            )

            # expiration_time used to be 'max_age' but the dogpile setting
            # is expiration_time. Support max_age for backwards compat.
            self._cache_expiration_time = cache_settings.get(
                'expiration_time',
                cache_settings.get('max_age', self._cache_expiration_time),
            )

            # If cache class is given, use that. If not, but if cache time
            # is given, default to memory. Otherwise, default to nothing.
            if self._cache_expiration_time:
                self._cache_class = 'dogpile.cache.memory'
            self._cache_class = self.cloud_config['cache'].get(
                'class', self._cache_class
            )

            self._cache_path = os.path.expanduser(
                cache_settings.get('path', self._cache_path)
            )
            self._cache_arguments = cache_settings.get(
                'arguments', self._cache_arguments
            )
            self._cache_expirations = cache_settings.get(
                'expiration', self._cache_expirations
            )

        if load_yaml_config:
            metrics_config = self.cloud_config.get('metrics', {})
            statsd_config = metrics_config.get('statsd', {})
            statsd_host = statsd_host or statsd_config.get('host')
            statsd_port = statsd_port or statsd_config.get('port')
            statsd_prefix = statsd_prefix or statsd_config.get('prefix')

            influxdb_cfg = metrics_config.get('influxdb', {})
            # Parse InfluxDB configuration
            if not influxdb_config:
                influxdb_config = influxdb_cfg
            else:
                influxdb_config.update(influxdb_cfg)

        if influxdb_config:
            config = {}
            if 'use_udp' in influxdb_config:
                use_udp = influxdb_config['use_udp']
                if isinstance(use_udp, str):
                    use_udp = use_udp.lower() in ('true', 'yes', '1')
                elif not isinstance(use_udp, bool):
                    use_udp = False
                    self.log.warning(
                        'InfluxDB.use_udp value type is not '
                        'supported. Use one of '
                        '[true|false|yes|no|1|0]'
                    )
                config['use_udp'] = use_udp
            for key in [
                'host',
                'port',
                'username',
                'password',
                'database',
                'measurement',
                'timeout',
            ]:
                if key in influxdb_config:
                    config[key] = influxdb_config[key]
            self._influxdb_config = config

        if load_envvars:
            statsd_host = statsd_host or os.environ.get('STATSD_HOST')
            statsd_port = statsd_port or os.environ.get('STATSD_PORT')
            statsd_prefix = statsd_prefix or os.environ.get('STATSD_PREFIX')

        self._statsd_host = statsd_host
        self._statsd_port = statsd_port
        self._statsd_prefix = statsd_prefix

        # Flag location to hold the peeked value of an argparse timeout value
        self._argv_timeout = False

        # Save the password callback
        # password = self._pw_callback(prompt="Password: ")
        self._pw_callback = pw_func

    def _get_os_environ(
        self, envvar_prefix: str | None = None
    ) -> dict[str, ty.Any] | None:
        ret = self._defaults_module.get_defaults()
        if not envvar_prefix:
            # This makes the or below be OS_ or OS_ which is a no-op
            envvar_prefix = 'OS_'
        environkeys = [
            k
            for k in os.environ.keys()
            if (k.startswith('OS_') or k.startswith(envvar_prefix))
            and not k.startswith('OS_TEST')  # infra CI var
            and not k.startswith('OS_STD')  # oslotest var
            and not k.startswith('OS_LOG')  # oslotest var
        ]
        for k in environkeys:
            newkey = k.split('_', 1)[-1].lower()
            ret[newkey] = os.environ[k]
        # If the only environ keys are selectors or behavior modification,
        # don't return anything
        selectors = {
            'OS_CLOUD',
            'OS_REGION_NAME',
            'OS_CLIENT_CONFIG_FILE',
            'OS_CLIENT_SECURE_FILE',
            'OS_CLOUD_NAME',
        }
        if set(environkeys) - selectors:
            return ret
        return None

    def _get_envvar(self, key: str, default: str | None = None) -> str | None:
        if not self._load_envvars:
            return default
        return os.environ.get(key, default)

    def get_extra_config(
        self, key: str, defaults: dict[str, ty.Any] | None = None
    ) -> dict[str, ty.Any]:
        """Fetch an arbitrary extra chunk of config, laying in defaults.

        :param string key: name of the config section to fetch
        :param dict defaults: (optional) default values to merge under the
                                         found config
        """
        defaults = _util.normalize_keys(defaults or {})
        assert defaults is not None  # narrow type
        if not key:
            return defaults
        return _util.merge_clouds(
            defaults, _util.normalize_keys(self.cloud_config.get(key, {}))
        )

    def _load_config_file(
        self,
    ) -> tuple[str, dict[str, ty.Any]] | tuple[None, None]:
        return self._load_yaml_json_file(self._config_files)

    def _load_secure_file(
        self,
    ) -> tuple[str, dict[str, ty.Any]] | tuple[None, None]:
        return self._load_yaml_json_file(self._secure_files)

    def _load_vendor_file(
        self,
    ) -> tuple[str, dict[str, ty.Any]] | tuple[None, None]:
        return self._load_yaml_json_file(self._vendor_files)

    def _load_yaml_json_file(
        self, filelist: list[str]
    ) -> tuple[str, dict[str, ty.Any]] | tuple[None, None]:
        for path in filelist:
            if os.path.exists(path):
                try:
                    with open(path) as f:
                        if path.endswith('json'):
                            return path, json.load(f)
                        else:
                            return path, yaml.safe_load(f)
                except OSError as e:
                    if e.errno == errno.EACCES:
                        # Can't access file so let's continue to the next
                        # file
                        continue
        return (None, None)

    def _validate_config_file(self, path: str, data: ty.Any) -> bool:
        """Validate config file contains a clouds entry.

        All config files should have a 'clouds' key at a minimum.
        """
        if not isinstance(data, dict):
            raise exceptions.ConfigException(
                f'Configuration file {path} is empty or not a valid mapping'
            )

        if 'clouds' not in data:
            # TODO(stephenfin): This should probably be an error at some point
            self.log.warning(
                "Configuration file %s does not contain a 'clouds' key", path
            )
            return False

        return True

    def _expand_region_name(self, region_name: str) -> dict[str, ty.Any]:
        return {'name': region_name, 'values': {}}

    def _expand_regions(
        self, regions: list[str | dict[str, ty.Any]]
    ) -> list[dict[str, ty.Any]]:
        ret = []
        for region in regions:
            if isinstance(region, dict):
                # i.e. must have name key, and only name,values keys
                if 'name' not in region or not {'name', 'values'} >= set(
                    region
                ):
                    raise exceptions.ConfigException(
                        f'Invalid region entry at: {region}'
                    )
                if 'values' not in region:
                    region['values'] = {}
                ret.append(copy.deepcopy(region))
            else:
                ret.append(self._expand_region_name(region))
        return ret

    def _get_regions(self, cloud: str) -> list[dict[str, ty.Any]]:
        if cloud not in self.cloud_config['clouds']:
            return [self._expand_region_name('')]
        regions = self._get_known_regions(cloud)
        if not regions:
            # We don't know of any regions use a workable default.
            regions = [self._expand_region_name('')]
        return regions

    def _get_known_regions(self, cloud: str) -> list[dict[str, ty.Any]]:
        config = _util.normalize_keys(self.cloud_config['clouds'][cloud])
        if 'regions' in config:
            return self._expand_regions(config['regions'])
        elif 'region_name' in config:
            if isinstance(config['region_name'], list):
                regions = config['region_name']
            else:
                regions = config['region_name'].split(',')
            if len(regions) > 1:
                warnings.warn(
                    f"Comma separated lists in region_name are deprecated. "
                    f"Please use a yaml list in the regions "
                    f"parameter in {self.config_filename} instead.",
                    os_warnings.OpenStackDeprecationWarning,
                )
            return self._expand_regions(regions)
        else:
            # crappit. we don't have a region defined.
            new_cloud: dict[str, ty.Any] = {}
            our_cloud = self.cloud_config['clouds'].get(cloud, {})
            self._expand_vendor_profile(cloud, new_cloud, our_cloud)
            if 'regions' in new_cloud and new_cloud['regions']:
                return self._expand_regions(new_cloud['regions'])
            elif 'region_name' in new_cloud and new_cloud['region_name']:
                return [self._expand_region_name(new_cloud['region_name'])]

        return []

    def _get_region(
        self, cloud: str | None = None, region_name: str = ''
    ) -> dict[str, ty.Any]:
        if region_name is None:
            region_name = ''

        if not cloud:
            return self._expand_region_name(region_name)

        regions = self._get_known_regions(cloud)
        if not regions:
            return self._expand_region_name(region_name)

        if not region_name:
            return regions[0]

        for region in regions:
            if region['name'] == region_name:
                return region

        raise exceptions.ConfigException(
            'Region {region_name} is not a valid region name for cloud '
            '{cloud}. Valid choices are {region_list}. Please note that '
            'region names are case sensitive.'.format(
                region_name=region_name,
                region_list=','.join([r['name'] for r in regions]),
                cloud=cloud,
            )
        )

    def get_cloud_names(self) -> list[str]:
        return list(self.cloud_config['clouds'].keys())

    def _get_base_cloud_config(
        self, name: str | None, profile: str | None = None
    ) -> dict[str, ty.Any]:
        cloud = {}

        # Only validate cloud name if one was given
        if name and name not in self.cloud_config['clouds']:
            raise exceptions.ConfigException(f"Cloud {name} was not found.")

        our_cloud = self.cloud_config['clouds'].get(name, dict())
        if profile:
            our_cloud['profile'] = profile

        # Get the defaults
        cloud.update(self.defaults)
        self._expand_vendor_profile(name, cloud, our_cloud)

        if 'auth' not in cloud:
            cloud['auth'] = {}

        _auth_update(cloud, our_cloud)
        if 'cloud' in cloud:
            del cloud['cloud']

        return cloud

    def _expand_vendor_profile(
        self,
        name: str | None,
        cloud: dict[str, ty.Any],
        our_cloud: dict[str, ty.Any],
    ) -> None:
        # Expand a profile if it exists. 'cloud' is an old confusing name
        # for this.
        profile_name = our_cloud.get('profile', our_cloud.get('cloud', None))
        if not profile_name or profile_name == self.envvar_key:
            return
        if 'cloud' in our_cloud:
            warnings.warn(
                f"{self.config_filename} uses the keyword 'cloud' to "
                f"reference a known vendor profile. This has been deprecated "
                f"in favor of the 'profile' keyword.",
                os_warnings.OpenStackDeprecationWarning,
            )

        vendor_filename, vendor_file = self._load_vendor_file()
        if (
            vendor_file
            and 'public-clouds' in vendor_file
            and profile_name in vendor_file['public-clouds']
        ):
            _auth_update(cloud, vendor_file['public-clouds'][profile_name])
        else:
            profile_data = vendors.get_profile(profile_name)
            if profile_data:
                nested_profile = profile_data.pop('profile', None)
                if nested_profile:
                    nested_profile_data = vendors.get_profile(nested_profile)
                    if nested_profile_data:
                        profile_data = nested_profile_data
                status = profile_data.pop('status', 'active')
                message = profile_data.pop('message', '')
                if status == 'deprecated':
                    warnings.warn(
                        f"{profile_name} is deprecated: {message}",
                        os_warnings.OpenStackDeprecationWarning,
                    )
                elif status == 'shutdown':
                    raise exceptions.ConfigException(
                        f"{profile_name} references a cloud that no longer "
                        f"exists: {message}"
                    )
                _auth_update(cloud, profile_data)
            else:
                # Can't find the requested vendor config, go about business
                warnings.warn(
                    f"Couldn't find the vendor profile {profile_name} for "
                    f"the cloud {name}",
                    os_warnings.ConfigurationWarning,
                )

    def _project_scoped(self, cloud: dict[str, ty.Any]) -> bool:
        return (
            'project_id' in cloud
            or 'project_name' in cloud
            or 'project_id' in cloud['auth']
            or 'project_name' in cloud['auth']
        )

    def _validate_networks(
        self, networks: list[dict[str, ty.Any]], key: str
    ) -> None:
        value = None
        for net in networks:
            if value and net[key]:
                raise exceptions.ConfigException(
                    "Duplicate network entries for {key}: {net1} and {net2}. "
                    "Only one network can be flagged with {key}".format(
                        key=key, net1=value['name'], net2=net['name']
                    )
                )
            if not value and net[key]:
                value = net

    def _fix_backwards_networks(
        self, cloud: dict[str, ty.Any]
    ) -> dict[str, ty.Any]:
        # Leave the external_network and internal_network keys in the
        # dict because consuming code might be expecting them.
        networks = []
        # Normalize existing network entries
        for net in cloud.get('networks', []):
            name = net.get('name')
            if not name:
                raise exceptions.ConfigException(
                    'Entry in network list is missing required field "name".'
                )
            network = dict(
                name=name,
                routes_externally=get_boolean(net.get('routes_externally')),
                nat_source=get_boolean(net.get('nat_source')),
                nat_destination=get_boolean(net.get('nat_destination')),
                default_interface=get_boolean(net.get('default_interface')),
            )
            # routes_ipv4_externally defaults to the value of routes_externally
            network['routes_ipv4_externally'] = get_boolean(
                net.get('routes_ipv4_externally', network['routes_externally'])
            )
            # routes_ipv6_externally defaults to the value of routes_externally
            network['routes_ipv6_externally'] = get_boolean(
                net.get('routes_ipv6_externally', network['routes_externally'])
            )
            networks.append(network)

        for key in ('external_network', 'internal_network'):
            external = key.startswith('external')
            if key in cloud and 'networks' in cloud:
                raise exceptions.ConfigException(
                    f"Both {key} and networks were specified in the config. "
                    f"Please remove {key} from the config and use the network "
                    f"list to configure network behavior."
                )
            if key in cloud:
                warnings.warn(
                    f"{key} is deprecated. Please replace with an entry in "
                    f"a dict inside of the networks list with name: "
                    f"{cloud[key]} and routes_externally: {external}",
                    os_warnings.OpenStackDeprecationWarning,
                )
                networks.append(
                    dict(
                        name=cloud[key],
                        routes_externally=external,
                        nat_destination=not external,
                        default_interface=external,
                    )
                )

        # Validate that we don't have duplicates
        self._validate_networks(networks, 'nat_destination')
        self._validate_networks(networks, 'default_interface')

        cloud['networks'] = networks
        return cloud

    def _handle_domain_id(self, cloud: dict[str, ty.Any]) -> dict[str, ty.Any]:
        # Allow people to just specify domain once if it's the same
        mappings = {
            'domain_id': ('user_domain_id', 'project_domain_id'),
            'domain_name': ('user_domain_name', 'project_domain_name'),
        }
        for target_key, possible_values in mappings.items():
            if not self._project_scoped(cloud):
                if target_key in cloud and target_key not in cloud['auth']:
                    cloud['auth'][target_key] = cloud.pop(target_key)
                continue
            for key in possible_values:
                if target_key in cloud['auth'] and key not in cloud['auth']:
                    cloud['auth'][key] = cloud['auth'][target_key]
            cloud.pop(target_key, None)
            cloud['auth'].pop(target_key, None)
        return cloud

    def _fix_backwards_auth(
        self, cloud: dict[str, ty.Any]
    ) -> dict[str, ty.Any]:
        mappings = {
            'domain_id': ('domain_id', 'domain-id'),
            'domain_name': ('domain_name', 'domain-name'),
            'user_domain_id': ('user_domain_id', 'user-domain-id'),
            'user_domain_name': ('user_domain_name', 'user-domain-name'),
            'project_domain_id': ('project_domain_id', 'project-domain-id'),
            'project_domain_name': (
                'project_domain_name',
                'project-domain-name',
            ),
            'token': ('auth-token', 'auth_token', 'token'),
            'passcode': ('passcode',),
        }
        if cloud.get('auth_type', None) == 'v2password':
            # If v2password is explcitly requested, this is to deal with old
            # clouds. That's fine - we need to map settings in the opposite
            # direction
            mappings['tenant_id'] = (
                'project_id',
                'project-id',
                'tenant_id',
                'tenant-id',
            )
            mappings['tenant_name'] = (
                'project_name',
                'project-name',
                'tenant_name',
                'tenant-name',
            )
        else:
            mappings['project_id'] = (
                'tenant_id',
                'tenant-id',
                'project_id',
                'project-id',
            )
            mappings['project_name'] = (
                'tenant_name',
                'tenant-name',
                'project_name',
                'project-name',
            )
        for target_key, possible_values in mappings.items():
            target = None
            for key in possible_values:
                if key in cloud['auth']:
                    target = str(cloud['auth'][key])
                    del cloud['auth'][key]
                # Prefer values NOT from the 'auth' section
                # as they may contain cli or environment overrides.
                # See story 2010784 for context.
                if key in cloud:
                    target = str(cloud[key])
                    del cloud[key]
            if target:
                cloud['auth'][target_key] = target
        return cloud

    def _fix_backwards_auth_plugin(
        self, cloud: dict[str, ty.Any]
    ) -> dict[str, ty.Any]:
        # Do the lists backwards so that auth_type is the ultimate winner
        mappings = {
            'auth_type': ('auth_plugin', 'auth_type'),
        }
        for target_key, possible_values in mappings.items():
            target = None
            for key in possible_values:
                if key in cloud:
                    target = cloud[key]
                    del cloud[key]
            cloud[target_key] = target
        # Because we force alignment to v3 nouns, we want to force
        # use of the auth plugin that can do auto-selection and dealing
        # with that based on auth parameters. v2password is basically
        # completely broken
        return cloud

    def register_argparse_arguments(
        self,
        parser: argparse_mod.ArgumentParser,
        argv: list[str],
        service_keys: list[str] | None = None,
    ) -> None:
        """Register all of the common argparse options needed.

        Given an argparse parser, register the keystoneauth Session arguments,
        the keystoneauth Auth Plugin Options and os-cloud. Also, peek in the
        argv to see if all of the auth plugin options should be registered
        or merely the ones already configured.

        :param argparse.ArgumentParser: parser to attach argparse options to
        :param argv: the arguments provided to the application
        :param string service_keys: Service or list of services this argparse
                                    should be specialized for, if known.
                                    The first item in the list will be used
                                    as the default value for service_type
                                    (optional)

        :raises exceptions.ConfigException if an invalid auth-type is requested
        """

        if service_keys is None:
            service_keys = []

        # Fix argv in place - mapping any keys with embedded _ in them to -
        _fix_argv(argv)

        local_parser = argparse_mod.ArgumentParser(add_help=False)

        for p in (parser, local_parser):
            p.add_argument(
                '--os-cloud',
                metavar='<name>',
                default=self._get_envvar('OS_CLOUD', None),
                help='Named cloud to connect to',
            )

        # we need to peek to see if timeout was actually passed, since
        # the keystoneauth declaration of it has a default, which means
        # we have no clue if the value we get is from the ksa default
        # for from the user passing it explicitly. We'll stash it for later
        local_parser.add_argument('--timeout', metavar='<timeout>')

        # We need for get_one to be able to peek at whether a token
        # was passed so that we can swap the default from password to
        # token if it was. And we need to also peek for --os-auth-token
        # for novaclient backwards compat
        local_parser.add_argument('--os-token')
        local_parser.add_argument('--os-auth-token')

        # Peek into the future and see if we have an auth-type set in
        # config AND a cloud set, so that we know which command line
        # arguments to register and show to the user (the user may want
        # to say something like:
        #   openstack --os-cloud=foo --os-oidctoken=bar
        # although I think that user is the cause of my personal pain
        options, _args = local_parser.parse_known_args(argv)
        if options.timeout:
            self._argv_timeout = True

        # validate = False because we're not _actually_ loading here
        # we're only peeking, so it's the wrong time to assert that
        # the rest of the arguments given are invalid for the plugin
        # chosen (for instance, --help may be requested, so that the
        # user can see what options he may want to give
        cloud_region = self.get_one(argparse=options, validate=False)
        default_auth_type = cloud_region.config['auth_type']

        try:
            loading.register_auth_argparse_arguments(
                parser, argv, default=default_auth_type
            )
        except Exception:
            # Hidiing the keystoneauth exception because we're not actually
            # loading the auth plugin at this point, so the error message
            # from it doesn't actually make sense to os-client-config users
            options, _args = parser.parse_known_args(argv)
            plugin_names = loading.get_available_plugin_names()
            raise exceptions.ConfigException(
                "An invalid auth-type was specified: {auth_type}. "
                "Valid choices are: {plugin_names}.".format(
                    auth_type=options.os_auth_type,
                    plugin_names=",".join(plugin_names),
                )
            )

        if service_keys:
            primary_service = service_keys[0]
        else:
            primary_service = None
        loading.register_session_argparse_arguments(parser)
        adapter.register_adapter_argparse_arguments(
            parser, service_type=primary_service
        )
        for service_key in service_keys:
            # legacy clients have un-prefixed api-version options
            parser.add_argument(
                '--{service_key}-api-version'.format(
                    service_key=service_key.replace('_', '-')
                ),
                help=argparse_mod.SUPPRESS,
            )
            adapter.register_service_adapter_argparse_arguments(
                parser, service_type=service_key
            )

        # Backwards compat options for legacy clients
        parser.add_argument('--http-timeout', help=argparse_mod.SUPPRESS)
        parser.add_argument('--os-endpoint-type', help=argparse_mod.SUPPRESS)
        parser.add_argument('--endpoint-type', help=argparse_mod.SUPPRESS)

    def _fix_backwards_interface(
        self, cloud: dict[str, ty.Any]
    ) -> dict[str, ty.Any]:
        new_cloud = {}
        for key in cloud.keys():
            if key.endswith('endpoint_type'):
                target_key = key.replace('endpoint_type', 'interface')
            else:
                target_key = key
            new_cloud[target_key] = cloud[key]
        return new_cloud

    def _fix_backwards_api_timeout(
        self, cloud: dict[str, ty.Any]
    ) -> dict[str, ty.Any]:
        new_cloud = {}
        # requests can only have one timeout, which means that in a single
        # cloud there is no point in different timeout values. However,
        # for some reason many of the legacy clients decided to shove their
        # service name in to the arg name for reasons surpassin sanity. If
        # we find any values that are not api_timeout, overwrite api_timeout
        # with the value
        service_timeout = None
        for key in cloud.keys():
            if key.endswith('timeout') and not (
                key == 'timeout' or key == 'api_timeout'
            ):
                service_timeout = cloud[key]
            else:
                new_cloud[key] = cloud[key]
        if service_timeout is not None:
            new_cloud['api_timeout'] = service_timeout
        # The common argparse arg from keystoneauth is called timeout, but
        # os-client-config expects it to be called api_timeout
        if self._argv_timeout:
            if 'timeout' in new_cloud and new_cloud['timeout']:
                new_cloud['api_timeout'] = new_cloud.pop('timeout')
        return new_cloud

    def get_all(self) -> list[cloud_region.CloudRegion]:
        clouds = []

        for cloud in self.get_cloud_names():
            for region in self._get_regions(cloud):
                if region:
                    clouds.append(
                        self.get_one(cloud, region_name=region['name'])
                    )
        return clouds

    def get_all_clouds(self) -> list[cloud_region.CloudRegion]:
        warnings.warn(
            "The 'get_all_clouds' method is a deprecated alias for "
            "'get_clouds' and will be removed in a future release.",
            os_warnings.RemovedInSDK60Warning,
        )
        return self.get_all()

    def _fix_args(
        self,
        args: dict[str, ty.Any] | None = None,
        argparse: argparse_mod.Namespace | None = None,
    ) -> dict[str, ty.Any]:
        """Massage the passed-in options

        Replace - with _ and strip os_ prefixes.

        Convert an argparse Namespace object to a dict, removing values
        that are either None or ''.
        """
        if not args:
            args = {}

        if argparse:
            # Convert the passed-in Namespace
            o_dict = vars(argparse)
            parsed_args = dict()
            for k in o_dict:
                if o_dict[k] is not None and o_dict[k] != '':
                    parsed_args[k] = o_dict[k]
            args.update(parsed_args)

        os_args = {}
        new_args = {}
        for key, val in iter(args.items()):
            if isinstance(args[key], dict):
                # dive into the auth dict
                new_args[key] = self._fix_args(args[key])
                continue

            key = key.replace('-', '_')
            if key.startswith('os_'):
                os_args[key[3:]] = val
            else:
                new_args[key] = val
        new_args.update(os_args)
        return new_args

    def _find_winning_auth_value(
        self, opt: 'opts.Opt', config: dict[str, dict[str, ty.Any]]
    ) -> dict[str, ty.Any] | None:
        opt_name = opt.name.replace('-', '_')
        if opt_name in config:
            return config[opt_name]
        else:
            deprecated = getattr(
                opt, 'deprecated', getattr(opt, 'deprecated_opts', [])
            )
            for d_opt in deprecated:
                d_opt_name = d_opt.name.replace('-', '_')
                if d_opt_name in config:
                    return config[d_opt_name]

        return None

    def auth_config_hook(self, config: dict[str, ty.Any]) -> dict[str, ty.Any]:
        """Allow examination of config values before loading auth plugin

        OpenStackClient will override this to perform additional checks
        on auth_type.
        """
        return config

    def _get_auth_loader(
        self, config: dict[str, ty.Any]
    ) -> loading.BaseLoader[ty.Any]:
        # Use the 'none' plugin for variants of None specified,
        # since it does not look up endpoints or tokens but rather
        # does a passthrough. This is useful for things like Ironic
        # that have a keystoneless operational mode, but means we're
        # still dealing with a keystoneauth Session object, so all the
        # _other_ things (SSL arg handling, timeout) all work consistently
        if config['auth_type'] in (None, "None", ''):
            config['auth_type'] = 'none'
        elif config['auth_type'] == 'token_endpoint':
            # Humans have been trained to use a thing called token_endpoint
            # That it does not exist in keystoneauth is irrelvant- it not
            # doing what they want causes them sorrow.
            config['auth_type'] = 'admin_token'

        loader: loading.BaseLoader[ty.Any] = loading.get_plugin_loader(
            config['auth_type']
        )

        # As the name would suggest, v3multifactor uses multiple factors for
        # authentication. As a result, we need to register the configuration
        # options for each required auth method. Normally, this is handled by
        # the 'MultiFactor.load_from_options' method but there doesn't appear
        # to be a way to "register" the auth methods without actually loading
        # the plugin. As a result, if we encounter this auth type then we need
        # to do this registration of extra options manually.
        # FIXME(stephenfin): We need to provide a mechanism to extend the
        # options in keystoneauth1.loading._plugins.identity.v3.MultiAuth
        # without calling 'load_from_options'.
        if config['auth_type'] == 'v3multifactor':
            if ty.TYPE_CHECKING:
                # narrow types
                assert isinstance(loader, v3_loaders.MultiFactor)
            # We use '.get' since we can't be sure this key is set yet -
            # validation happens later, in _validate_auth
            loader._methods = config.get('auth_methods')

        return loader

    def _validate_auth(
        self, config: dict[str, ty.Any], loader: loading.BaseLoader[ty.Any]
    ) -> dict[str, ty.Any]:
        # May throw a keystoneauth1.exceptions.NoMatchingPlugin

        plugin_options = loader.get_options()

        for p_opt in plugin_options:
            # if it's in config.auth, win, kill it from config dict
            # if it's in config and not in config.auth, move it
            # deprecated loses to current
            # provided beats default, deprecated or not
            winning_value = self._find_winning_auth_value(
                p_opt,
                config['auth'],
            )
            if not winning_value:
                winning_value = self._find_winning_auth_value(
                    p_opt,
                    config,
                )

            config = self._clean_up_after_ourselves(
                config,
                p_opt,
                winning_value,
            )

            if winning_value:
                # Prefer the plugin configuration dest value if the value's key
                # is marked as deprecated.
                if p_opt.dest is None:
                    good_name = p_opt.name.replace('-', '_')
                    config['auth'][good_name] = winning_value
                else:
                    config['auth'][p_opt.dest] = winning_value

            # See if this needs a prompting
            config = self.option_prompt(config, p_opt)

        return config

    def _validate_auth_correctly(
        self, config: dict[str, ty.Any], loader: loading.BaseLoader[ty.Any]
    ) -> dict[str, ty.Any]:
        # May throw a keystoneauth1.exceptions.NoMatchingPlugin

        plugin_options = loader.get_options()

        for p_opt in plugin_options:
            # if it's in config, win, move it and kill it from config dict
            # if it's in config.auth but not in config it's good
            # deprecated loses to current
            # provided beats default, deprecated or not
            winning_value = self._find_winning_auth_value(
                p_opt,
                config,
            )
            if not winning_value:
                winning_value = self._find_winning_auth_value(
                    p_opt,
                    config['auth'],
                )

            config = self._clean_up_after_ourselves(
                config,
                p_opt,
                winning_value,
            )

            # See if this needs a prompting
            config = self.option_prompt(config, p_opt)

        return config

    def option_prompt(
        self, config: dict[str, ty.Any], p_opt: 'opts.Opt'
    ) -> dict[str, ty.Any]:
        """Prompt user for option that requires a value"""
        if (
            getattr(p_opt, 'prompt', None) is not None
            and p_opt.dest not in config['auth']
            and self._pw_callback is not None
        ):
            config['auth'][p_opt.dest] = self._pw_callback(p_opt.prompt)
        return config

    def _clean_up_after_ourselves(
        self,
        config: dict[str, ty.Any],
        p_opt: 'opts.Opt',
        winning_value: ty.Any,
    ) -> dict[str, ty.Any]:
        # Clean up after ourselves
        for opt in [p_opt.name] + [o.name for o in p_opt.deprecated]:
            opt = opt.replace('-', '_')
            config.pop(opt, None)
            config['auth'].pop(opt, None)

        if winning_value:
            # Prefer the plugin configuration dest value if the value's key
            # is marked as depreciated.
            if p_opt.dest is None:
                config['auth'][p_opt.name.replace('-', '_')] = winning_value
            else:
                config['auth'][p_opt.dest] = winning_value
        return config

    def _handle_value_types(
        self, config: dict[str, ty.Any]
    ) -> dict[str, ty.Any]:
        for key in BOOL_KEYS:
            if key in config:
                if not isinstance(config[key], bool):
                    config[key] = get_boolean(config[key])

        for key in CSV_KEYS:
            if key in config:
                if isinstance(config[key], str):
                    config[key] = config[key].split(',')
        return config

    def magic_fixes(self, config: dict[str, ty.Any]) -> dict[str, ty.Any]:
        """Perform the set of magic argument fixups"""

        # These backwards compat values are only set via argparse. If it's
        # there, it's because it was passed in explicitly, and should win
        config = self._fix_backwards_api_timeout(config)
        config = self._fix_backwards_auth_plugin(config)
        config = self._fix_backwards_auth(config)
        config = self._fix_backwards_interface(config)
        config = self._fix_backwards_networks(config)
        config = self._handle_domain_id(config)
        config = self._handle_value_types(config)

        # TODO(mordred): Special casing auth_url here. We should
        #                come back to this betterer later so that it's
        #                more generalized
        if 'auth' in config and 'auth_url' in config['auth']:
            config['auth']['auth_url'] = config['auth']['auth_url'].format(
                **config
            )

        return config

    def get_one(
        self,
        cloud: str | None = None,
        validate: bool = True,
        argparse: argparse_mod.Namespace | None = None,
        **kwargs: ty.Any,
    ) -> cloud_region.CloudRegion:
        """Retrieve a single CloudRegion and merge additional options

        :param string cloud:
            The name of the configuration to load from clouds.yaml
        :param boolean validate:
            Validate the config. Setting this to False causes no auth plugin
            to be created. It's really only useful for testing.
        :param Namespace argparse:
            An argparse Namespace object; allows direct passing in of
            argparse options to be added to the cloud config.  Values
            of None and '' will be removed.
        :param region_name: Name of the region of the cloud.
        :param kwargs: Additional configuration options

        :returns: openstack.config.cloud_region.CloudRegion
        :raises: keystoneauth1.exceptions.MissingRequiredOptions
            on missing required auth parameters
        """

        profile = kwargs.pop('profile', None)
        args = self._fix_args(kwargs, argparse=argparse)

        if cloud is None:
            if 'cloud' in args:
                cloud = args['cloud']
            else:
                cloud = self.default_cloud

        config = self._get_base_cloud_config(cloud, profile)

        # Get region specific settings
        if 'region_name' not in args:
            args['region_name'] = ''
        region = self._get_region(cloud=cloud, region_name=args['region_name'])
        args['region_name'] = region['name']
        region_args = copy.deepcopy(region['values'])

        # Regions is a list that we can use to create a list of cloud/region
        # objects. It does not belong in the single-cloud dict
        config.pop('regions', None)

        # Can't just do update, because None values take over
        for arg_list in region_args, args:
            for key, val in iter(arg_list.items()):
                if val is not None:
                    if key == 'auth' and config[key] is not None:
                        config[key] = _auth_update(config[key], val)
                    else:
                        config[key] = val

        config = self.magic_fixes(config)
        config = _util.normalize_keys(config)

        # NOTE(dtroyer): OSC needs a hook into the auth args before the
        #                plugin is loaded in order to maintain backward-
        #                compatible behaviour
        config = self.auth_config_hook(config)

        if validate:
            loader = self._get_auth_loader(config)
            config = self._validate_auth(config, loader)
            auth_plugin = loader.load_from_options(**config['auth'])
        else:
            auth_plugin = None

        # If any of the defaults reference other values, we need to expand
        for key, value in config.items():
            if hasattr(value, 'format') and key not in FORMAT_EXCLUSIONS:
                config[key] = value.format(**config)

        force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
        prefer_ipv6 = config.pop('prefer_ipv6', True)
        if not prefer_ipv6:
            force_ipv4 = True

        # Override global metrics config with more specific per-cloud
        # details.
        metrics_config = config.get('metrics', {})
        statsd_config = metrics_config.get('statsd', {})
        statsd_host = statsd_config.get('host') or self._statsd_host
        statsd_port = statsd_config.get('port') or self._statsd_port
        statsd_prefix = statsd_config.get('prefix') or self._statsd_prefix
        influxdb_config = metrics_config.get('influxdb', {})
        if influxdb_config:
            merged_influxdb = copy.deepcopy(self._influxdb_config)
            merged_influxdb.update(influxdb_config)
            influxdb_config = merged_influxdb
        else:
            influxdb_config = self._influxdb_config

        if cloud is None:
            cloud_name = ''
        else:
            cloud_name = str(cloud)
        return self._cloud_region_class(
            name=cloud_name,
            region_name=config['region_name'],
            config=config,
            extra_config=self.extra_config,
            force_ipv4=force_ipv4,
            auth_plugin=auth_plugin,
            openstack_config=self,
            session_constructor=self._session_constructor,
            app_name=self._app_name,
            app_version=self._app_version,
            cache_auth=self._cache_auth,
            cache_expiration_time=self._cache_expiration_time,
            cache_expirations=self._cache_expirations,
            cache_path=self._cache_path,
            cache_class=self._cache_class,
            cache_arguments=self._cache_arguments,
            password_callback=self._pw_callback,
            statsd_host=statsd_host,
            statsd_port=statsd_port,
            statsd_prefix=statsd_prefix,
            influxdb_config=influxdb_config,
        )

    def get_one_cloud(
        self,
        cloud: str | None = None,
        validate: bool = True,
        argparse: argparse_mod.Namespace | None = None,
        **kwargs: ty.Any,
    ) -> cloud_region.CloudRegion:
        warnings.warn(
            "The 'get_one_cloud' method is a deprecated alias for 'get_one' "
            "and will be removed in a future release.",
            os_warnings.RemovedInSDK60Warning,
        )
        return self.get_one(
            cloud=cloud,
            validate=validate,
            argparse=argparse,
            **kwargs,
        )

    def get_one_cloud_osc(
        self,
        cloud: str | None = None,
        validate: bool = True,
        argparse: argparse_mod.Namespace | None = None,
        **kwargs: ty.Any,
    ) -> cloud_region.CloudRegion:
        """Retrieve a single CloudRegion and merge additional options

        :param string cloud:
            The name of the configuration to load from clouds.yaml
        :param boolean validate:
            Validate the config. Setting this to False causes no auth plugin
            to be created. It's really only useful for testing.
        :param Namespace argparse:
            An argparse Namespace object; allows direct passing in of
            argparse options to be added to the cloud config.  Values
            of None and '' will be removed.
        :param region_name: Name of the region of the cloud.
        :param kwargs: Additional configuration options

        :raises: keystoneauth1.exceptions.MissingRequiredOptions
            on missing required auth parameters
        """

        args = self._fix_args(kwargs, argparse=argparse)

        if cloud is None:
            if 'cloud' in args:
                cloud = args['cloud']
            else:
                cloud = self.default_cloud

        config = self._get_base_cloud_config(cloud)

        # Get region specific settings
        if 'region_name' not in args:
            args['region_name'] = ''
        region = self._get_region(cloud=cloud, region_name=args['region_name'])
        args['region_name'] = region['name']
        region_args = copy.deepcopy(region['values'])

        # Regions is a list that we can use to create a list of cloud/region
        # objects. It does not belong in the single-cloud dict
        config.pop('regions', None)

        # Can't just do update, because None values take over
        for arg_list in region_args, args:
            for key, val in iter(arg_list.items()):
                if val is not None:
                    if key == 'auth' and config[key] is not None:
                        config[key] = _auth_update(config[key], val)
                    else:
                        config[key] = val

        config = self.magic_fixes(config)

        # NOTE(dtroyer): OSC needs a hook into the auth args before the
        #                plugin is loaded in order to maintain backward-
        #                compatible behaviour
        config = self.auth_config_hook(config)

        if validate:
            loader = self._get_auth_loader(config)
            config = self._validate_auth_correctly(config, loader)
            auth_plugin = loader.load_from_options(**config['auth'])
        else:
            auth_plugin = None

        # If any of the defaults reference other values, we need to expand
        for key, value in config.items():
            if hasattr(value, 'format') and key not in FORMAT_EXCLUSIONS:
                config[key] = value.format(**config)

        force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
        prefer_ipv6 = config.pop('prefer_ipv6', True)
        if not prefer_ipv6:
            force_ipv4 = True

        if cloud is None:
            cloud_name = ''
        else:
            cloud_name = str(cloud)
        return self._cloud_region_class(
            name=cloud_name,
            region_name=config['region_name'],
            config=config,
            extra_config=self.extra_config,
            force_ipv4=force_ipv4,
            auth_plugin=auth_plugin,
            openstack_config=self,
            cache_auth=self._cache_auth,
            cache_expiration_time=self._cache_expiration_time,
            cache_expirations=self._cache_expirations,
            cache_path=self._cache_path,
            cache_class=self._cache_class,
            cache_arguments=self._cache_arguments,
            password_callback=self._pw_callback,
        )

    @staticmethod
    def set_one_cloud(
        config_file: str,
        cloud: str,
        set_config: dict[str, ty.Any] | None = None,
    ) -> None:
        """Set a single cloud configuration.

        :param string config_file:
            The path to the config file to edit. If this file does not exist
            it will be created.
        :param string cloud:
            The name of the configuration to save to clouds.yaml
        :param dict set_config: Configuration options to be set
        """

        set_config = set_config or {}
        cur_config = {}
        try:
            with open(config_file) as fh:
                cur_config = yaml.safe_load(fh)
        except OSError as e:
            # Not no such file
            if e.errno != 2:
                raise
            pass

        clouds_config = cur_config.get('clouds', {})
        cloud_config = _auth_update(clouds_config.get(cloud, {}), set_config)
        clouds_config[cloud] = cloud_config
        cur_config['clouds'] = clouds_config

        with open(config_file, 'w') as fh:
            yaml.safe_dump(cur_config, fh, default_flow_style=False)


if __name__ == '__main__':
    config = OpenStackConfig().get_all_clouds()
    for cloud in config:
        print_cloud = False
        if len(sys.argv) == 1:
            print_cloud = True
        elif len(sys.argv) == 3 and (
            sys.argv[1] == cloud.name and sys.argv[2] == cloud.region
        ):
            print_cloud = True
        elif len(sys.argv) == 2 and (sys.argv[1] == cloud.name):
            print_cloud = True

        if print_cloud:
            print(cloud.name, cloud.region, cloud.config)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import limit
from openstack.tests.unit import base

EXAMPLE = {
    "totalShareNetworksUsed": 0,
    "maxTotalShareGigabytes": 1000,
    "maxTotalShareNetworks": 10,
    "totalSharesUsed": 0,
    "totalShareGigabytesUsed": 0,
    "totalShareSnapshotsUsed": 0,
    "maxTotalShares": 50,
    "totalSnapshotGigabytesUsed": 0,
    "maxTotalSnapshotGigabytes": 1000,
    "maxTotalShareSnapshots": 50,
    "maxTotalShareReplicas": 100,
    "maxTotalReplicaGigabytes": 1000,
    "totalShareReplicasUsed": 0,
    "totalReplicaGigabytesUsed": 0,
}


class TestLimit(base.TestCase):
    def test_basic(self):
        limits = limit.Limit()
        self.assertEqual('limits', limits.resources_key)
        self.assertEqual('/limits', limits.base_path)
        self.assertTrue(limits.allow_list)
        self.assertFalse(limits.allow_fetch)
        self.assertFalse(limits.allow_create)
        self.assertFalse(limits.allow_commit)
        self.assertFalse(limits.allow_delete)
        self.assertFalse(limits.allow_head)

    def test_make_limits(self):
        limits = limit.Limit(**EXAMPLE)
        self.assertEqual(
            EXAMPLE['totalShareNetworksUsed'], limits.totalShareNetworksUsed
        )
        self.assertEqual(
            EXAMPLE['maxTotalShareGigabytes'], limits.maxTotalShareGigabytes
        )
        self.assertEqual(
            EXAMPLE['maxTotalShareNetworks'], limits.maxTotalShareNetworks
        )
        self.assertEqual(EXAMPLE['totalSharesUsed'], limits.totalSharesUsed)
        self.assertEqual(
            EXAMPLE['totalShareGigabytesUsed'], limits.totalShareGigabytesUsed
        )
        self.assertEqual(
            EXAMPLE['totalShareSnapshotsUsed'], limits.totalShareSnapshotsUsed
        )
        self.assertEqual(EXAMPLE['maxTotalShares'], limits.maxTotalShares)
        self.assertEqual(
            EXAMPLE['totalSnapshotGigabytesUsed'],
            limits.totalSnapshotGigabytesUsed,
        )
        self.assertEqual(
            EXAMPLE['maxTotalSnapshotGigabytes'],
            limits.maxTotalSnapshotGigabytes,
        )
        self.assertEqual(
            EXAMPLE['maxTotalShareSnapshots'], limits.maxTotalShareSnapshots
        )
        self.assertEqual(
            EXAMPLE['maxTotalShareReplicas'], limits.maxTotalShareReplicas
        )
        self.assertEqual(
            EXAMPLE['maxTotalReplicaGigabytes'],
            limits.maxTotalReplicaGigabytes,
        )
        self.assertEqual(
            EXAMPLE['totalShareReplicasUsed'], limits.totalShareReplicasUsed
        )
        self.assertEqual(
            EXAMPLE['totalReplicaGigabytesUsed'],
            limits.totalReplicaGigabytesUsed,
        )



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_inventory
----------------------------------

Functional tests for inventory methods.
"""

from openstack.cloud import inventory
from openstack.tests.functional import base


class TestInventory(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        if not self.operator_cloud:
            self.skipTest("Operator cloud is required for this test")

        # This needs to use an admin account, otherwise a public IP
        # is not allocated from devstack.
        self.inventory = inventory.OpenStackInventory(cloud='devstack-admin')
        self.server_name = self.getUniqueString('inventory')
        self.addCleanup(self._cleanup_server)
        server = self.operator_cloud.create_server(
            name=self.server_name,
            image=self.image,
            flavor=self.flavor,
            wait=True,
            auto_ip=True,
            network='public',
        )
        self.server_id = server['id']

    def _cleanup_server(self):
        self.user_cloud.delete_server(self.server_id, wait=True)

    def _test_host_content(self, host):
        self.assertEqual(host['image']['id'], self.image.id)
        self.assertIsInstance(host['volumes'], list)
        self.assertIsInstance(host['metadata'], dict)
        self.assertIn('interface_ip', host)
        self.assertIn('ram', host['flavor'])

    def _test_expanded_host_content(self, host):
        self.assertEqual(host['image']['name'], self.image.name)
        self.assertEqual(host['flavor']['name'], self.flavor.name)

    def test_get_host(self):
        host = self.inventory.get_host(self.server_id)
        self.assertIsNotNone(host)
        self.assertEqual(host['name'], self.server_name)
        self._test_host_content(host)
        self._test_expanded_host_content(host)
        host_found = False
        for host in self.inventory.list_hosts():
            if host['id'] == self.server_id:
                host_found = True
                self._test_host_content(host)
        self.assertTrue(host_found)

    def test_get_host_no_detail(self):
        host = self.inventory.get_host(self.server_id, expand=False)
        self.assertIsNotNone(host)
        self.assertEqual(host['name'], self.server_name)

        self.assertEqual(host['image']['id'], self.image.id)
        self.assertNotIn('links', host['image'])
        self.assertNotIn('name', host['name'])
        self.assertIn('ram', host['flavor'])

        host_found = False
        for host in self.inventory.list_hosts(expand=False):
            if host['id'] == self.server_id:
                host_found = True
                self._test_host_content(host)
        self.assertTrue(host_found)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_object
----------------------------------

Functional tests for object methods.
"""

import random
import string
import tempfile

from testtools import content

from openstack.tests.functional import base


class TestObject(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        if not self.user_cloud.has_service('object-store'):
            self.skipTest('Object service not supported by cloud')

    def test_create_object(self):
        '''Test uploading small and large files.'''
        container_name = self.getUniqueString('container')
        self.addDetail('container', content.text_content(container_name))
        self.addCleanup(self.user_cloud.delete_container, container_name)
        self.user_cloud.create_container(container_name)
        container = self.user_cloud.get_container(container_name)
        self.assertEqual(container_name, container.name)
        self.assertEqual(
            [], self.user_cloud.list_containers(prefix='somethin')
        )
        sizes = (
            (64 * 1024, 1),  # 64K, one segment
            (64 * 1024, 5),  # 64MB, 5 segments
        )
        for size, nseg in sizes:
            segment_size = int(round(size / nseg))
            with tempfile.NamedTemporaryFile() as fake_file:
                fake_content = ''.join(
                    random.SystemRandom().choice(
                        string.ascii_uppercase + string.digits
                    )
                    for _ in range(size)
                ).encode('latin-1')

                fake_file.write(fake_content)
                fake_file.flush()
                name = f'test-{size}'
                self.addCleanup(
                    self.user_cloud.delete_object, container_name, name
                )
                self.user_cloud.create_object(
                    container_name,
                    name,
                    fake_file.name,
                    segment_size=segment_size,
                    metadata={'foo': 'bar'},
                )
                self.assertFalse(
                    self.user_cloud.is_object_stale(
                        container_name, name, fake_file.name
                    )
                )
            self.assertEqual(
                'bar',
                self.user_cloud.get_object_metadata(container_name, name)[
                    'foo'
                ],
            )
            self.user_cloud.update_object(
                container=container_name,
                name=name,
                metadata={'testk': 'testv'},
            )
            self.assertEqual(
                'testv',
                self.user_cloud.get_object_metadata(container_name, name)[
                    'testk'
                ],
            )
            self.assertIsNotNone(
                self.user_cloud.get_object(container_name, name)
            )
            self.assertEqual(
                name, self.user_cloud.list_objects(container_name)[0]['name']
            )
            self.assertEqual(
                [], self.user_cloud.list_objects(container_name, prefix='abc')
            )
            self.assertTrue(
                self.user_cloud.delete_object(container_name, name)
            )
        self.assertEqual([], self.user_cloud.list_objects(container_name))
        self.assertEqual(
            container_name, self.user_cloud.get_container(container_name).name
        )
        self.user_cloud.delete_container(container_name)

    def test_download_object_to_file(self):
        '''Test uploading small and large files.'''
        container_name = self.getUniqueString('container')
        self.addDetail('container', content.text_content(container_name))
        self.addCleanup(self.user_cloud.delete_container, container_name)
        self.user_cloud.create_container(container_name)
        self.assertEqual(
            container_name, self.user_cloud.list_containers()[0]['name']
        )
        sizes = (
            (64 * 1024, 1),  # 64K, one segment
            (64 * 1024, 5),  # 64MB, 5 segments
        )
        for size, nseg in sizes:
            fake_content = b''
            segment_size = int(round(size / nseg))
            with tempfile.NamedTemporaryFile() as fake_file:
                fake_content = ''.join(
                    random.SystemRandom().choice(
                        string.ascii_uppercase + string.digits
                    )
                    for _ in range(size)
                ).encode('latin-1')

                fake_file.write(fake_content)
                fake_file.flush()
                name = f'test-{size}'
                self.addCleanup(
                    self.user_cloud.delete_object, container_name, name
                )
                self.user_cloud.create_object(
                    container_name,
                    name,
                    fake_file.name,
                    segment_size=segment_size,
                    metadata={'foo': 'bar'},
                )
                self.assertFalse(
                    self.user_cloud.is_object_stale(
                        container_name, name, fake_file.name
                    )
                )
            self.assertEqual(
                'bar',
                self.user_cloud.get_object_metadata(container_name, name)[
                    'foo'
                ],
            )
            self.user_cloud.update_object(
                container=container_name,
                name=name,
                metadata={'testk': 'testv'},
            )
            self.assertEqual(
                'testv',
                self.user_cloud.get_object_metadata(container_name, name)[
                    'testk'
                ],
            )

            with tempfile.NamedTemporaryFile() as fake_file:
                self.user_cloud.get_object(
                    container_name, name, outfile=fake_file.name
                )
                downloaded_content = open(fake_file.name, 'rb').read()
                self.assertEqual(fake_content, downloaded_content)

            self.assertEqual(
                name, self.user_cloud.list_objects(container_name)[0]['name']
            )
            self.assertTrue(
                self.user_cloud.delete_object(container_name, name)
            )
        self.assertEqual([], self.user_cloud.list_objects(container_name))
        self.assertEqual(
            container_name, self.user_cloud.list_containers()[0]['name']
        )
        self.user_cloud.delete_container(container_name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack import resource


class Fake(resource.Resource):
    resource_key = "resource"
    resources_key = "resources"
    base_path = "/fake"

    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_head = True

    #: The transaction date and time.
    timestamp = resource.Header("x-timestamp")
    #: The name of this resource.
    name = resource.Body("name", alternate_id=True)
    #: The value of the resource. Also available in headers.
    value = resource.Body("value", alias="x-resource-value")
    #: Is this resource cool? If so, set it to True.
    #: This is a multi-line comment about cool stuff.
    cool = resource.Body("cool", type=bool)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.message.v2 import _proxy
from openstack.message.v2 import claim
from openstack.message.v2 import message
from openstack.message.v2 import queue
from openstack.message.v2 import subscription
from openstack import proxy as proxy_base
from openstack.tests.unit import test_proxy_base

QUEUE_NAME = 'test_queue'


class TestMessageProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestMessageQueue(TestMessageProxy):
    def test_queue_create(self):
        self.verify_create(self.proxy.create_queue, queue.Queue)

    def test_queue_get(self):
        self.verify_get(self.proxy.get_queue, queue.Queue)
        self.verify_get_overrided(
            self.proxy, queue.Queue, 'openstack.message.v2.queue.Queue'
        )

    def test_queues(self):
        self.verify_list(self.proxy.queues, queue.Queue)

    def test_queue_delete(self):
        self.verify_delete(self.proxy.delete_queue, queue.Queue, False)

    def test_queue_delete_ignore(self):
        self.verify_delete(self.proxy.delete_queue, queue.Queue, True)


class TestMessageMessage(TestMessageProxy):
    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_message_post(self, mock_get_resource):
        message_obj = message.Message(queue_name="test_queue")
        mock_get_resource.return_value = message_obj
        self._verify(
            "openstack.message.v2.message.Message.post",
            self.proxy.post_message,
            method_args=["test_queue", ["msg1", "msg2"]],
            expected_args=[self.proxy, ["msg1", "msg2"]],
        )
        mock_get_resource.assert_called_once_with(
            message.Message, None, queue_name="test_queue"
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_message_get(self, mock_get_resource):
        mock_get_resource.return_value = "resource_or_id"
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_message,
            method_args=["test_queue", "resource_or_id"],
            expected_args=[message.Message, "resource_or_id"],
        )
        mock_get_resource.assert_called_once_with(
            message.Message, "resource_or_id", queue_name="test_queue"
        )
        self.verify_get_overrided(
            self.proxy, message.Message, 'openstack.message.v2.message.Message'
        )

    def test_messages(self):
        self.verify_list(
            self.proxy.messages,
            message.Message,
            method_kwargs={"queue_name": "test_queue"},
            expected_kwargs={"queue_name": "test_queue"},
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_message_delete(self, mock_get_resource):
        fake_message = mock.Mock()
        fake_message.id = "message_id"
        mock_get_resource.return_value = fake_message
        self._verify(
            "openstack.proxy.Proxy._delete",
            self.proxy.delete_message,
            method_args=["test_queue", "resource_or_id", None, False],
            expected_args=[message.Message, fake_message],
            expected_kwargs={"ignore_missing": False},
        )
        self.assertIsNone(fake_message.claim_id)
        mock_get_resource.assert_called_once_with(
            message.Message, "resource_or_id", queue_name="test_queue"
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_message_delete_claimed(self, mock_get_resource):
        fake_message = mock.Mock()
        fake_message.id = "message_id"
        mock_get_resource.return_value = fake_message
        self._verify(
            "openstack.proxy.Proxy._delete",
            self.proxy.delete_message,
            method_args=["test_queue", "resource_or_id", "claim_id", False],
            expected_args=[message.Message, fake_message],
            expected_kwargs={"ignore_missing": False},
        )
        self.assertEqual("claim_id", fake_message.claim_id)
        mock_get_resource.assert_called_once_with(
            message.Message, "resource_or_id", queue_name="test_queue"
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_message_delete_ignore(self, mock_get_resource):
        fake_message = mock.Mock()
        fake_message.id = "message_id"
        mock_get_resource.return_value = fake_message
        self._verify(
            "openstack.proxy.Proxy._delete",
            self.proxy.delete_message,
            method_args=["test_queue", "resource_or_id", None, True],
            expected_args=[message.Message, fake_message],
            expected_kwargs={"ignore_missing": True},
        )
        self.assertIsNone(fake_message.claim_id)
        mock_get_resource.assert_called_once_with(
            message.Message, "resource_or_id", queue_name="test_queue"
        )


class TestMessageSubscription(TestMessageProxy):
    def test_subscription_create(self):
        self._verify(
            "openstack.message.v2.subscription.Subscription.create",
            self.proxy.create_subscription,
            method_args=["test_queue"],
            expected_args=[self.proxy],
            expected_kwargs={"base_path": None},
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_subscription_get(self, mock_get_resource):
        mock_get_resource.return_value = "resource_or_id"
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_subscription,
            method_args=["test_queue", "resource_or_id"],
            expected_args=[subscription.Subscription, "resource_or_id"],
        )
        mock_get_resource.assert_called_once_with(
            subscription.Subscription,
            "resource_or_id",
            queue_name="test_queue",
        )
        self.verify_get_overrided(
            self.proxy,
            subscription.Subscription,
            'openstack.message.v2.subscription.Subscription',
        )

    def test_subscriptions(self):
        self.verify_list(
            self.proxy.subscriptions,
            subscription.Subscription,
            method_kwargs={"queue_name": "test_queue"},
            expected_kwargs={"queue_name": "test_queue"},
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_subscription_delete(self, mock_get_resource):
        mock_get_resource.return_value = "test_subscription"
        self.verify_delete(
            self.proxy.delete_subscription,
            subscription.Subscription,
            ignore_missing=False,
            method_args=["test_queue", "resource_or_id"],
            expected_args=["test_subscription"],
        )
        mock_get_resource.assert_called_once_with(
            subscription.Subscription,
            "resource_or_id",
            queue_name="test_queue",
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_subscription_delete_ignore(self, mock_get_resource):
        mock_get_resource.return_value = "test_subscription"
        self.verify_delete(
            self.proxy.delete_subscription,
            subscription.Subscription,
            ignore_missing=True,
            method_args=["test_queue", "resource_or_id"],
            expected_args=["test_subscription"],
        )
        mock_get_resource.assert_called_once_with(
            subscription.Subscription,
            "resource_or_id",
            queue_name="test_queue",
        )


class TestMessageClaim(TestMessageProxy):
    def test_claim_create(self):
        self._verify(
            "openstack.message.v2.claim.Claim.create",
            self.proxy.create_claim,
            method_args=["test_queue"],
            expected_args=[self.proxy],
            expected_kwargs={"base_path": None},
        )

    def test_claim_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_claim,
            method_args=["test_queue", "resource_or_id"],
            expected_args=[claim.Claim, "resource_or_id"],
            expected_kwargs={"queue_name": "test_queue"},
        )
        self.verify_get_overrided(
            self.proxy, claim.Claim, 'openstack.message.v2.claim.Claim'
        )

    def test_claim_update(self):
        self._verify(
            "openstack.proxy.Proxy._update",
            self.proxy.update_claim,
            method_args=["test_queue", "resource_or_id"],
            method_kwargs={"k1": "v1"},
            expected_args=[claim.Claim, "resource_or_id"],
            expected_kwargs={"queue_name": "test_queue", "k1": "v1"},
        )

    def test_claim_delete(self):
        self.verify_delete(
            self.proxy.delete_claim,
            claim.Claim,
            ignore_missing=False,
            method_args=["test_queue", "test_claim"],
            expected_args=["test_claim"],
            expected_kwargs={
                "queue_name": "test_queue",
                "ignore_missing": False,
            },
        )

    def test_claim_delete_ignore(self):
        self.verify_delete(
            self.proxy.delete_claim,
            claim.Claim,
            ignore_missing=True,
            method_args=["test_queue", "test_claim"],
            expected_args=["test_claim"],
            expected_kwargs={
                "queue_name": "test_queue",
                "ignore_missing": True,
            },
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import testtools

from openstack import exceptions
from openstack.tests.unit import base


class TestUsers(base.TestCase):
    def _get_keystone_mock_url(
        self, resource, append=None, v3=True, qs_elements=None
    ):
        base_url_append = None
        if v3:
            base_url_append = 'v3'
        return self.get_mock_url(
            service_type='identity',
            resource=resource,
            append=append,
            base_url_append=base_url_append,
            qs_elements=qs_elements,
        )

    def _get_user_list(self, user_data):
        uri = self._get_keystone_mock_url(resource='users')
        return {
            'users': [
                user_data.json_response['user'],
            ],
            'links': {
                'self': uri,
                'previous': None,
                'next': None,
            },
        }

    def test_create_user_v3(self):
        user_data = self._get_user_data(
            domain_id=uuid.uuid4().hex,
            description=self.getUniqueString('description'),
        )

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self._get_keystone_mock_url(resource='users'),
                    status_code=200,
                    json=user_data.json_response,
                    validate=dict(json=user_data.json_request),
                ),
            ]
        )

        user = self.cloud.create_user(
            name=user_data.name,
            email=user_data.email,
            password=user_data.password,
            description=user_data.description,
            domain_id=user_data.domain_id,
        )

        self.assertEqual(user_data.name, user.name)
        self.assertEqual(user_data.email, user.email)
        self.assertEqual(user_data.description, user.description)
        self.assertEqual(user_data.user_id, user.id)
        self.assert_calls()

    def test_create_user_v3_no_domain(self):
        user_data = self._get_user_data(
            domain_id=uuid.uuid4().hex, email='test@example.com'
        )
        with testtools.ExpectedException(
            exceptions.SDKException,
            "User or project creation requires an explicit "
            "domain_id argument.",
        ):
            self.cloud.create_user(
                name=user_data.name,
                email=user_data.email,
                password=user_data.password,
            )

    def test_delete_user(self):
        user_data = self._get_user_data(domain_id=uuid.uuid4().hex)
        user_resource_uri = self._get_keystone_mock_url(
            resource='users', append=[user_data.user_id]
        )

        self.register_uris(
            [
                # you can't lookup by name, so return 404 for that attempt
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='users', append=[user_data.name]
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='users',
                        qs_elements=[f'name={user_data.name}'],
                    ),
                    status_code=200,
                    json=self._get_user_list(user_data),
                ),
                dict(method='DELETE', uri=user_resource_uri, status_code=204),
            ]
        )

        self.cloud.delete_user(user_data.name)
        self.assert_calls()

    def test_delete_user_not_found(self):
        nonexistent_user_id = self.getUniqueString()
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='users', append=[nonexistent_user_id]
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(resource='users'),
                    status_code=200,
                    json={'users': []},
                ),
            ]
        )
        self.assertFalse(self.cloud.delete_user(nonexistent_user_id))

    def test_add_user_to_group(self):
        user_data = self._get_user_data()
        group_data = self._get_group_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='users', append=[user_data.user_id]
                    ),
                    status_code=200,
                    json=user_data.json_response,
                ),
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='groups', append=[group_data.group_id]
                    ),
                    status_code=200,
                    json=group_data.json_response,
                ),
                dict(
                    method='PUT',
                    uri=self._get_keystone_mock_url(
                        resource='groups',
                        append=[
                            group_data.group_id,
                            'users',
                            user_data.user_id,
                        ],
                    ),
                    status_code=200,
                ),
            ]
        )
        self.cloud.add_user_to_group(user_data.user_id, group_data.group_id)
        self.assert_calls()

    def test_is_user_in_group(self):
        user_data = self._get_user_data()
        group_data = self._get_group_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='users', append=[user_data.user_id]
                    ),
                    status_code=200,
                    json=user_data.json_response,
                ),
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='groups', append=[group_data.group_id]
                    ),
                    status_code=200,
                    json=group_data.json_response,
                ),
                dict(
                    method='HEAD',
                    uri=self._get_keystone_mock_url(
                        resource='groups',
                        append=[
                            group_data.group_id,
                            'users',
                            user_data.user_id,
                        ],
                    ),
                    status_code=204,
                ),
            ]
        )

        self.assertTrue(
            self.cloud.is_user_in_group(user_data.user_id, group_data.group_id)
        )
        self.assert_calls()

    def test_remove_user_from_group(self):
        user_data = self._get_user_data()
        group_data = self._get_group_data()

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='users', append=[user_data.user_id]
                    ),
                    json=user_data.json_response,
                ),
                dict(
                    method='GET',
                    uri=self._get_keystone_mock_url(
                        resource='groups', append=[group_data.group_id]
                    ),
                    status_code=200,
                    json=group_data.json_response,
                ),
                dict(
                    method='DELETE',
                    uri=self._get_keystone_mock_url(
                        resource='groups',
                        append=[
                            group_data.group_id,
                            'users',
                            user_data.user_id,
                        ],
                    ),
                    status_code=204,
                ),
            ]
        )

        self.cloud.remove_user_from_group(
            user_data.user_id, group_data.group_id
        )
        self.assert_calls()



# Copyright(c) 2018 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack import resource


class Segment(resource.Resource):
    resource_key = "segment"
    resources_key = "segments"
    base_path = "/segments"

    # capabilities
    # 1] GET /v1/segments
    # 2] GET /v1/segments/<segment_uuid>
    # 3] POST /v1/segments
    # 4] PUT /v1/segments/<segment_uuid>
    # 5] DELETE /v1/segments/<segment_uuid>
    allow_list = True
    allow_fetch = True
    allow_create = True
    allow_commit = True
    allow_delete = True

    # add enabled flag to segment in 1.2
    _max_microversion = '1.2'

    #: A ID of representing this segment.
    id = resource.Body("id")
    #: A Uuid of representing this segment.
    uuid = resource.Body("uuid")
    #: A created time of representing this segment.
    created_at = resource.Body("created_at")
    #: A latest updated time of representing this segment.
    updated_at = resource.Body("updated_at")
    #: The name of this segment.
    name = resource.Body("name")
    #: The description of this segment.
    description = resource.Body("description")
    #: The recovery method of this segment.
    recovery_method = resource.Body("recovery_method")
    #: The service type of this segment.
    service_type = resource.Body("service_type")
    #: The enabled flag of this segment.
    is_enabled = resource.Body("enabled", type=bool)

    _query_mapping = resource.QueryParameters(
        "sort_key",
        "sort_dir",
        recovery_method="recovery_method",
        service_type="service_type",
        is_enabled="enabled",
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack import resource
from openstack import utils


class Transfer(resource.Resource):
    resource_key = "transfer"
    resources_key = "transfers"
    base_path = "/os-volume-transfer"

    # capabilities
    allow_create = True
    allow_delete = True
    allow_fetch = True
    allow_list = True

    # Properties
    #: UUID of the transfer.
    id = resource.Body("id")
    #: The date and time when the resource was created.
    created_at = resource.Body("created_at")
    #: Name of the volume to transfer.
    name = resource.Body("name")
    #: ID of the volume to transfer.
    volume_id = resource.Body("volume_id")
    #: Auth key for the transfer.
    auth_key = resource.Body("auth_key")
    #: A list of links associated with this volume. *Type: list*
    links = resource.Body("links")

    def accept(self, session, *, auth_key=None):
        """Accept a volume transfer.

        :param session: The session to use for making this request.
        :param auth_key: The authentication key for the volume transfer.

        :return: This :class:`Transfer` instance.
        """
        body = {'accept': {'auth_key': auth_key}}

        url = utils.urljoin(self.base_path, self.id, 'accept')
        resp = session.post(url, json=body)
        exceptions.raise_from_response(resp)

        transfer = Transfer()
        transfer._translate_response(resp)
        return transfer



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Managing policies in the Cluster service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/clustering.html
"""

EVENT_ID = "5d982071-76c5-4733-bf35-b9e38a563c99"


def list_events(conn):
    print("List Events:")

    for events in conn.clustering.events():
        print(events.to_dict())

    for events in conn.clustering.events(sort='name:asc'):
        print(events.to_dict())


def get_event(conn):
    print("Get Event:")

    event = conn.clustering.get_event(EVENT_ID)
    print(event.to_dict())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
List resources from the Identity service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/identity.html
"""


def list_users(conn):
    print("List Users:")

    for user in conn.identity.users():
        print(user)


def list_credentials(conn):
    print("List Credentials:")

    for credential in conn.identity.credentials():
        print(credential)


def list_projects(conn):
    print("List Projects:")

    for project in conn.identity.projects():
        print(project)


def list_domains(conn):
    print("List Domains:")

    for domain in conn.identity.domains():
        print(domain)


def list_groups(conn):
    print("List Groups:")

    for group in conn.identity.groups():
        print(group)


def list_services(conn):
    print("List Services:")

    for service in conn.identity.services():
        print(service)


def list_endpoints(conn):
    print("List Endpoints:")

    for endpoint in conn.identity.endpoints():
        print(endpoint)


def list_regions(conn):
    print("List Regions:")

    for region in conn.identity.regions():
        print(region)


def list_roles(conn):
    print("List Roles:")

    for role in conn.identity.roles():
        print(role)


def list_role_domain_group_assignments(conn):
    print("List Roles assignments for a group on domain:")

    for role in conn.identity.role_domain_group_assignments():
        print(role)


def list_role_domain_user_assignments(conn):
    print("List Roles assignments for a user on domain:")

    for role in conn.identity.role_project_user_assignments():
        print(role)


def list_role_project_group_assignments(conn):
    print("List Roles assignments for a group on project:")

    for role in conn.identity.role_project_group_assignments():
        print(role)


def list_role_project_user_assignments(conn):
    print("List Roles assignments for a user on project:")

    for role in conn.identity.role_project_user_assignments():
        print(role)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.database.v1 import database
from openstack.tests.unit import base


IDENTIFIER = 'NAME'
INSTANCE_ID = 'INSTANCE_ID'
EXAMPLE = {
    'character_set': '1',
    'collate': '2',
    'instance_id': INSTANCE_ID,
    'name': IDENTIFIER,
}


class TestDatabase(base.TestCase):
    def test_basic(self):
        sot = database.Database()
        self.assertEqual('database', sot.resource_key)
        self.assertEqual('databases', sot.resources_key)
        path = '/instances/%(instance_id)s/databases'
        self.assertEqual(path, sot.base_path)
        self.assertTrue(sot.allow_list)
        self.assertTrue(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertTrue(sot.allow_delete)

    def test_make_it(self):
        sot = database.Database(**EXAMPLE)
        self.assertEqual(IDENTIFIER, sot.id)
        self.assertEqual(EXAMPLE['character_set'], sot.character_set)
        self.assertEqual(EXAMPLE['collate'], sot.collate)
        self.assertEqual(EXAMPLE['instance_id'], sot.instance_id)
        self.assertEqual(IDENTIFIER, sot.name)
        self.assertEqual(IDENTIFIER, sot.id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class VpnIpsecPolicy(resource.Resource):
    resource_key = 'ipsecpolicy'
    resources_key = 'ipsecpolicies'
    base_path = '/vpn/ipsecpolicies'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'auth_algorithm',
        'description',
        'encapsulation_mode',
        'encryption_algorithm',
        'name',
        'pfs',
        'project_id',
        'phase1_negotiation_mode',
        'transform_protocol',
    )

    # Properties
    #: The authentication hash algorithm. Valid values are sha1,
    # sha256, sha384, sha512. The default is sha1.
    auth_algorithm = resource.Body('auth_algorithm')
    #: A human-readable description for the resource.
    # Default is an empty string.
    description = resource.Body('description')
    #: The encapsulation mode. A valid value is tunnel or transport
    encapsulation_mode = resource.Body('encapsulation_mode')
    #: The encryption algorithm. A valid value is 3des, aes-128,
    # aes-192, aes-256, and so on. Default is aes-128.
    encryption_algorithm = resource.Body('encryption_algorithm')
    #: The lifetime of the security association. The lifetime consists
    # of a unit and integer value. You can omit either the unit or value
    # portion of the lifetime. Default unit is seconds and
    # default value is 3600.
    lifetime = resource.Body('lifetime', type=dict)
    #: Human-readable name of the resource. Default is an empty string.
    name = resource.Body('name')
    #: Perfect forward secrecy (PFS). A valid value is Group2,
    # Group5, Group14, and so on. Default is Group5.
    pfs = resource.Body('pfs')
    #: The ID of the project.
    project_id = resource.Body('project_id')
    #: The IKE mode. A valid value is main, which is the default.
    phase1_negotiation_mode = resource.Body('phase1_negotiation_mode')
    #: The transform protocol. A valid value is ESP, AH, or AH- ESP.
    transform_protocol = resource.Body('transform_protocol')
    #: The units for the lifetime of the security association.
    # The lifetime consists of a unit and integer value.
    # You can omit either the unit or value portion of the lifetime.
    # Default unit is seconds and default value is 3600.
    units = resource.Body('units')
    #: The lifetime value, as a positive integer. The lifetime
    # consists of a unit and integer value.
    # You can omit either the unit or value portion of the lifetime.
    # Default unit is seconds and default value is 3600.
    value = resource.Body('value', type=int)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.placement.v1 import resource_provider as _resource_provider
from openstack.tests.functional import base


class TestResourceProvider(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()

        if not self.operator_cloud.has_service('placement'):
            self.skipTest('placement service not supported by cloud')

        self.resource_provider_name = self.getUniqueString()

        resource_provider = (
            self.operator_cloud.placement.create_resource_provider(
                name=self.resource_provider_name,
            )
        )
        self.assertIsInstance(
            resource_provider, _resource_provider.ResourceProvider
        )
        self.assertEqual(self.resource_provider_name, resource_provider.name)
        self.resource_provider = resource_provider

    def tearDown(self):
        result = self.operator_cloud.placement.delete_resource_provider(
            self.resource_provider,
        )
        self.assertIsNone(result)
        super().tearDown()

    def test_resource_provider(self):
        # list all resource providers

        resource_providers = list(
            self.operator_cloud.placement.resource_providers()
        )
        self.assertIsInstance(
            resource_providers[0],
            _resource_provider.ResourceProvider,
        )
        self.assertIn(
            self.resource_provider_name,
            {x.name for x in resource_providers},
        )

        # retrieve details of the resource provider by name

        resource_provider = (
            self.operator_cloud.placement.find_resource_provider(
                self.resource_provider.name,
            )
        )
        self.assertEqual(self.resource_provider_name, resource_provider.name)

        # retrieve details of the resource provider by ID

        resource_provider = (
            self.operator_cloud.placement.get_resource_provider(
                self.resource_provider.id,
            )
        )
        self.assertEqual(self.resource_provider_name, resource_provider.name)

        # update the resource provider

        new_resource_provider_name = self.getUniqueString()

        resource_provider = (
            self.operator_cloud.placement.update_resource_provider(
                self.resource_provider,
                name=new_resource_provider_name,
                generation=self.resource_provider.generation,
            )
        )
        self.assertIsInstance(
            resource_provider,
            _resource_provider.ResourceProvider,
        )
        self.assertEqual(
            new_resource_provider_name,
            resource_provider.name,
        )

    def test_resource_provider_aggregates(self):
        aggregates = [uuid.uuid4().hex, uuid.uuid4().hex]

        # update the resource provider aggregates

        resource_provider = (
            self.operator_cloud.placement.set_resource_provider_aggregates(
                self.resource_provider,
                *aggregates,
            )
        )
        self.assertCountEqual(aggregates, resource_provider.aggregates)

        # retrieve details of resource provider aggregates

        resource_provider = (
            self.operator_cloud.placement.get_resource_provider_aggregates(
                self.resource_provider,
            )
        )
        self.assertCountEqual(aggregates, resource_provider.aggregates)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.dns.v2 import _base
from openstack import resource


class ZoneNameserver(_base.Resource):
    """DNS Zone Nameserver resource"""

    resources_key = 'nameservers'
    base_path = '/zones/%(zone_id)s/nameservers'

    # capabilities
    allow_list = True

    _query_mapping = resource.QueryParameters(
        include_pagination_defaults=False,
    )

    #: ID for the zone
    zone_id = resource.URI('zone_id')
    #: The hostname of the nameserver
    #: *Type: str*
    hostname = resource.Body('hostname')
    #: The priority of the nameserver
    #: *Type: int*
    priority = resource.Body('priority')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class PolicyType(resource.Resource):
    resource_key = 'policy_type'
    resources_key = 'policy_types'
    base_path = '/policy-types'

    # Capabilities
    allow_list = True
    allow_fetch = True

    # Properties
    #: Name of policy type.
    name = resource.Body('name', alternate_id=True)
    #: The schema of the policy type.
    schema = resource.Body('schema')
    #: The support status of the policy type
    support_status = resource.Body('support_status')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.shared_file_system.v2 import _proxy
from openstack.shared_file_system.v2 import limit
from openstack.shared_file_system.v2 import resource_locks
from openstack.shared_file_system.v2 import share
from openstack.shared_file_system.v2 import share_access_rule
from openstack.shared_file_system.v2 import share_group
from openstack.shared_file_system.v2 import share_group_snapshot
from openstack.shared_file_system.v2 import share_instance
from openstack.shared_file_system.v2 import share_network
from openstack.shared_file_system.v2 import share_network_subnet
from openstack.shared_file_system.v2 import share_snapshot
from openstack.shared_file_system.v2 import share_snapshot_instance
from openstack.shared_file_system.v2 import storage_pool
from openstack.shared_file_system.v2 import user_message
from openstack.tests.unit import test_proxy_base


class TestSharedFileSystemProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestSharedFileSystemShare(TestSharedFileSystemProxy):
    def test_shares(self):
        self.verify_list(self.proxy.shares, share.Share)

    def test_shares_detailed(self):
        self.verify_list(
            self.proxy.shares,
            share.Share,
            method_kwargs={"details": True, "query": 1},
            expected_kwargs={"query": 1},
        )

    def test_shares_not_detailed(self):
        self.verify_list(
            self.proxy.shares,
            share.Share,
            method_kwargs={"details": False, "query": 1},
            expected_kwargs={"query": 1},
        )

    def test_share_get(self):
        self.verify_get(self.proxy.get_share, share.Share)

    def test_share_find(self):
        self.verify_find(self.proxy.find_share, share.Share)

    def test_share_delete(self):
        self.verify_delete(self.proxy.delete_share, share.Share, False)

    def test_share_delete_ignore(self):
        self.verify_delete(self.proxy.delete_share, share.Share, True)

    def test_share_create(self):
        self.verify_create(self.proxy.create_share, share.Share)

    def test_share_update(self):
        self.verify_update(self.proxy.update_share, share.Share)

    def test_share_resize_extend(self):
        mock_share = share.Share(size=10, id='fakeId')
        self.proxy._get = mock.Mock(return_value=mock_share)

        self._verify(
            "openstack.shared_file_system.v2.share." + "Share.extend_share",
            self.proxy.resize_share,
            method_args=['fakeId', 20],
            expected_args=[self.proxy, 20, False],
        )

    def test_share_resize_shrink(self):
        mock_share = share.Share(size=30, id='fakeId')
        self.proxy._get = mock.Mock(return_value=mock_share)

        self._verify(
            "openstack.shared_file_system.v2.share." + "Share.shrink_share",
            self.proxy.resize_share,
            method_args=['fakeId', 20],
            expected_args=[self.proxy, 20],
        )

    def test_share_instances(self):
        self.verify_list(
            self.proxy.share_instances, share_instance.ShareInstance
        )

    def test_share_instance_get(self):
        self.verify_get(
            self.proxy.get_share_instance, share_instance.ShareInstance
        )

    def test_share_instance_reset(self):
        self._verify(
            "openstack.shared_file_system.v2.share_instance."
            + "ShareInstance.reset_status",
            self.proxy.reset_share_instance_status,
            method_args=['id', 'available'],
            expected_args=[self.proxy, 'available'],
        )

    def test_share_instance_delete(self):
        self._verify(
            "openstack.shared_file_system.v2.share_instance."
            + "ShareInstance.force_delete",
            self.proxy.delete_share_instance,
            method_args=['id'],
            expected_args=[self.proxy],
        )

    @mock.patch("openstack.resource.wait_for_status")
    def test_wait_for(self, mock_wait):
        mock_resource = mock.Mock()
        mock_wait.return_value = mock_resource

        self.proxy.wait_for_status(mock_resource, 'ACTIVE')

        mock_wait.assert_called_once_with(
            self.proxy, mock_resource, 'ACTIVE', None, 2, None, 'status', None
        )


class TestSharedFileSystemStoragePool(TestSharedFileSystemProxy):
    def test_storage_pools(self):
        self.verify_list(self.proxy.storage_pools, storage_pool.StoragePool)

    def test_storage_pool_detailed(self):
        self.verify_list(
            self.proxy.storage_pools,
            storage_pool.StoragePool,
            method_kwargs={"details": True, "backend": "alpha"},
            expected_kwargs={"backend": "alpha"},
        )

    def test_storage_pool_not_detailed(self):
        self.verify_list(
            self.proxy.storage_pools,
            storage_pool.StoragePool,
            method_kwargs={"details": False, "backend": "alpha"},
            expected_kwargs={"backend": "alpha"},
        )


class TestSharedFileSystemShareMetadata(TestSharedFileSystemProxy):
    def test_get_share_metadata(self):
        self._verify(
            "openstack.shared_file_system.v2.share.Share.fetch_metadata",
            self.proxy.get_share_metadata,
            method_args=["share_id"],
            expected_args=[self.proxy],
            expected_result=share.Share(
                id="share_id", metadata={"key": "value"}
            ),
        )

    def test_get_share_metadata_item(self):
        self._verify(
            "openstack.shared_file_system.v2.share.Share.get_metadata_item",
            self.proxy.get_share_metadata_item,
            method_args=["share_id", "key"],
            expected_args=[self.proxy, "key"],
            expected_result=share.Share(
                id="share_id", metadata={"key": "value"}
            ),
        )

    def test_create_share_metadata(self):
        metadata = {"foo": "bar", "newFoo": "newBar"}
        self._verify(
            "openstack.shared_file_system.v2.share.Share.set_metadata",
            self.proxy.create_share_metadata,
            method_args=["share_id"],
            method_kwargs=metadata,
            expected_args=[self.proxy],
            expected_kwargs={"metadata": metadata},
            expected_result=share.Share(id="share_id", metadata=metadata),
        )

    def test_update_share_metadata(self):
        metadata = {"foo": "bar", "newFoo": "newBar"}
        replace = True
        self._verify(
            "openstack.shared_file_system.v2.share.Share.set_metadata",
            self.proxy.update_share_metadata,
            method_args=["share_id", metadata, replace],
            expected_args=[self.proxy],
            expected_kwargs={"metadata": metadata, "replace": replace},
            expected_result=share.Share(id="share_id", metadata=metadata),
        )

    def test_delete_share_metadata(self):
        self._verify(
            "openstack.shared_file_system.v2.share.Share.delete_metadata_item",
            self.proxy.delete_share_metadata,
            expected_result=None,
            method_args=["share_id", ["key"]],
            expected_args=[self.proxy, "key"],
        )


class TestUserMessageProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_user_messages(self):
        self.verify_list(self.proxy.user_messages, user_message.UserMessage)

    def test_user_messages_queried(self):
        self.verify_list(
            self.proxy.user_messages,
            user_message.UserMessage,
            method_kwargs={"action_id": "1"},
            expected_kwargs={"action_id": "1"},
        )

    def test_user_message_get(self):
        self.verify_get(self.proxy.get_user_message, user_message.UserMessage)

    def test_delete_user_message(self):
        self.verify_delete(
            self.proxy.delete_user_message, user_message.UserMessage, False
        )

    def test_delete_user_message_true(self):
        self.verify_delete(
            self.proxy.delete_user_message, user_message.UserMessage, True
        )

    def test_limit(self):
        self.verify_list(self.proxy.limits, limit.Limit)


class TestShareSnapshotResource(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_share_snapshots(self):
        self.verify_list(
            self.proxy.share_snapshots, share_snapshot.ShareSnapshot
        )

    def test_share_snapshots_detailed(self):
        self.verify_list(
            self.proxy.share_snapshots,
            share_snapshot.ShareSnapshot,
            method_kwargs={"details": True, "name": "my_snapshot"},
            expected_kwargs={"name": "my_snapshot"},
        )

    def test_share_snapshots_not_detailed(self):
        self.verify_list(
            self.proxy.share_snapshots,
            share_snapshot.ShareSnapshot,
            method_kwargs={"details": False, "name": "my_snapshot"},
            expected_kwargs={"name": "my_snapshot"},
        )

    def test_share_snapshot_get(self):
        self.verify_get(
            self.proxy.get_share_snapshot, share_snapshot.ShareSnapshot
        )

    def test_share_snapshot_delete(self):
        self.verify_delete(
            self.proxy.delete_share_snapshot,
            share_snapshot.ShareSnapshot,
            False,
        )

    def test_share_snapshot_delete_ignore(self):
        self.verify_delete(
            self.proxy.delete_share_snapshot,
            share_snapshot.ShareSnapshot,
            True,
        )

    def test_share_snapshot_create(self):
        self.verify_create(
            self.proxy.create_share_snapshot, share_snapshot.ShareSnapshot
        )

    def test_share_snapshot_update(self):
        self.verify_update(
            self.proxy.update_share_snapshot, share_snapshot.ShareSnapshot
        )

    @mock.patch("openstack.resource.wait_for_delete")
    def test_wait_for_delete(self, mock_wait):
        mock_resource = mock.Mock()
        mock_wait.return_value = mock_resource

        self.proxy.wait_for_delete(mock_resource)

        mock_wait.assert_called_once_with(
            self.proxy, mock_resource, 2, 120, None
        )


class TestShareSnapshotInstanceResource(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_share_snapshot_instances(self):
        self.verify_list(
            self.proxy.share_snapshot_instances,
            share_snapshot_instance.ShareSnapshotInstance,
        )

    def test_share_snapshot_instance_detailed(self):
        self.verify_list(
            self.proxy.share_snapshot_instances,
            share_snapshot_instance.ShareSnapshotInstance,
            method_kwargs={"details": True, "query": {'snapshot_id': 'fake'}},
            expected_kwargs={"query": {'snapshot_id': 'fake'}},
        )

    def test_share_snapshot_instance_not_detailed(self):
        self.verify_list(
            self.proxy.share_snapshot_instances,
            share_snapshot_instance.ShareSnapshotInstance,
            method_kwargs={"details": False, "query": {'snapshot_id': 'fake'}},
            expected_kwargs={"query": {'snapshot_id': 'fake'}},
        )

    def test_share_snapshot_instance_get(self):
        self.verify_get(
            self.proxy.get_share_snapshot_instance,
            share_snapshot_instance.ShareSnapshotInstance,
        )


class TestShareNetworkResource(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_share_networks(self):
        self.verify_list(self.proxy.share_networks, share_network.ShareNetwork)

    def test_share_networks_detailed(self):
        self.verify_list(
            self.proxy.share_networks,
            share_network.ShareNetwork,
            method_kwargs={"details": True, "name": "my_net"},
            expected_kwargs={"name": "my_net"},
        )

    def test_share_networks_not_detailed(self):
        self.verify_list(
            self.proxy.share_networks,
            share_network.ShareNetwork,
            method_kwargs={"details": False, "name": "my_net"},
            expected_kwargs={"name": "my_net"},
        )

    def test_share_network_get(self):
        self.verify_get(
            self.proxy.get_share_network, share_network.ShareNetwork
        )

    def test_share_network_delete(self):
        self.verify_delete(
            self.proxy.delete_share_network, share_network.ShareNetwork, False
        )

    def test_share_network_delete_ignore(self):
        self.verify_delete(
            self.proxy.delete_share_network, share_network.ShareNetwork, True
        )

    def test_share_network_create(self):
        self.verify_create(
            self.proxy.create_share_network, share_network.ShareNetwork
        )

    def test_share_network_update(self):
        self.verify_update(
            self.proxy.update_share_network, share_network.ShareNetwork
        )


class TestShareNetworkSubnetResource(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_share_network_subnets(self):
        self.verify_list(
            self.proxy.share_network_subnets,
            share_network_subnet.ShareNetworkSubnet,
            method_args=["test_share"],
            expected_args=[],
            expected_kwargs={"share_network_id": "test_share"},
        )

    def test_share_network_subnet_get(self):
        self.verify_get(
            self.proxy.get_share_network_subnet,
            share_network_subnet.ShareNetworkSubnet,
            method_args=["fake_network_id", "fake_sub_network_id"],
            expected_args=['fake_sub_network_id'],
            expected_kwargs={'share_network_id': 'fake_network_id'},
        )

    def test_share_network_subnet_create(self):
        self.verify_create(
            self.proxy.create_share_network_subnet,
            share_network_subnet.ShareNetworkSubnet,
            method_args=["fake_network_id"],
            method_kwargs={"p1": "v1"},
            expected_args=[],
            expected_kwargs={
                "share_network_id": "fake_network_id",
                "p1": "v1",
            },
        )

    def test_share_network_subnet_delete(self):
        self.verify_delete(
            self.proxy.delete_share_network_subnet,
            share_network_subnet.ShareNetworkSubnet,
            False,
            method_args=["fake_network_id", "fake_sub_network_id"],
            expected_args=["fake_sub_network_id"],
            expected_kwargs={'share_network_id': 'fake_network_id'},
        )


class TestAccessRuleProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_access_ruless(self):
        self.verify_list(
            self.proxy.access_rules,
            share_access_rule.ShareAccessRule,
            method_args=["test_share"],
            expected_args=[],
            expected_kwargs={"share_id": "test_share"},
        )

    def test_access_rules_get(self):
        self.verify_get(
            self.proxy.get_access_rule, share_access_rule.ShareAccessRule
        )

    def test_access_rules_create(self):
        self.verify_create(
            self.proxy.create_access_rule,
            share_access_rule.ShareAccessRule,
            method_args=["share_id"],
            expected_args=[],
        )

    def test_access_rules_delete(self):
        self._verify(
            "openstack.shared_file_system.v2.share_access_rule.ShareAccessRule.delete",
            self.proxy.delete_access_rule,
            method_args=[
                'access_id',
                'share_id',
            ],
            expected_args=[self.proxy],
            expected_kwargs={'unrestrict': False},
        )


class TestResourceLocksProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_list_resource_locks(self):
        self.verify_list(
            self.proxy.resource_locks, resource_locks.ResourceLock
        )

    def test_resource_lock_get(self):
        self.verify_get(
            self.proxy.get_resource_lock, resource_locks.ResourceLock
        )

    def test_resource_lock_delete(self):
        self.verify_delete(
            self.proxy.delete_resource_lock, resource_locks.ResourceLock, False
        )

    def test_resource_lock_delete_ignore(self):
        self.verify_delete(
            self.proxy.delete_resource_lock, resource_locks.ResourceLock, True
        )

    def test_resource_lock_create(self):
        self.verify_create(
            self.proxy.create_resource_lock, resource_locks.ResourceLock
        )

    def test_resource_lock_update(self):
        self.verify_update(
            self.proxy.update_resource_lock, resource_locks.ResourceLock
        )


class TestShareGroupResource(test_proxy_base.TestProxyBase):
    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_share_groups(self):
        self.verify_list(self.proxy.share_groups, share_group.ShareGroup)

    def test_share_groups_query(self):
        self.verify_list(
            self.proxy.share_groups,
            share_group.ShareGroup,
            method_kwargs={"query": 1},
            expected_kwargs={"query": 1},
        )

    def test_share_group_get(self):
        self.verify_get(self.proxy.get_share_group, share_group.ShareGroup)

    def test_share_group_find(self):
        self.verify_find(self.proxy.find_share_group, share_group.ShareGroup)

    def test_share_group_delete(self):
        self.verify_delete(
            self.proxy.delete_share_group, share_group.ShareGroup, False
        )

    def test_share_group_delete_ignore(self):
        self.verify_delete(
            self.proxy.delete_share_group, share_group.ShareGroup, True
        )

    def test_share_group_create(self):
        self.verify_create(
            self.proxy.create_share_group, share_group.ShareGroup
        )

    def test_share_group_update(self):
        self.verify_update(
            self.proxy.update_share_group, share_group.ShareGroup
        )

    def test_share_group_snapshots(self):
        self.verify_list(
            self.proxy.share_group_snapshots,
            share_group_snapshot.ShareGroupSnapshot,
        )

    def test_share_group_snapshot_get(self):
        self.verify_get(
            self.proxy.get_share_group_snapshot,
            share_group_snapshot.ShareGroupSnapshot,
        )

    def test_share_group_snapshot_update(self):
        self.verify_update(
            self.proxy.update_share_group_snapshot,
            share_group_snapshot.ShareGroupSnapshot,
        )

    def test_share_group_snapshot_delete(self):
        self.verify_delete(
            self.proxy.delete_share_group_snapshot,
            share_group_snapshot.ShareGroupSnapshot,
            False,
        )

    def test_share_group_snapshot_delete_ignore(self):
        self.verify_delete(
            self.proxy.delete_share_group_snapshot,
            share_group_snapshot.ShareGroupSnapshot,
            True,
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def list_share_metadata(conn, share_id):
    # Method returns the entire share with the metadata inside it.
    returned_share = conn.get_share_metadata(share_id)

    # Access metadata of share
    metadata = returned_share['metadata']

    print("List All Share Metadata:")
    for meta_key in metadata:
        print(f"{meta_key}={metadata[meta_key]}")


def get_share_metadata_item(conn, share_id, key):
    # Method returns the entire share with the metadata inside it.
    returned_share = conn.get_share_metadata_item(share_id, key)

    # Access metadata of share
    metadata = returned_share['metadata']

    print("Get share metadata item given item key and share id:")
    print(metadata[key])


def create_share_metadata(conn, share_id, metadata):
    # Method returns the entire share with the metadata inside it.
    created_share = conn.create_share_metadata(share_id, metadata)

    # Access metadata of share
    metadata = created_share['metadata']

    print("Metadata created for given share:")
    print(metadata)


def update_share_metadata(conn, share_id, metadata):
    # Method returns the entire share with the metadata inside it.
    updated_share = conn.update_share_metadata(share_id, metadata, True)

    # Access metadata of share
    metadata = updated_share['metadata']

    print("Updated metadata for given share:")
    print(metadata)


def delete_share_metadata(conn, share_id, keys):
    # Method doesn't return anything.
    conn.delete_share_metadata(share_id, keys)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.compute.v2 import keypair as _keypair
from openstack.tests.functional import base


class TestKeypair(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()

        # Keypairs can't have .'s in the name. Because why?
        self.keypair_name = self.getUniqueString().split('.')[-1]

    def _delete_keypair(self, keypair):
        ret = self.user_cloud.compute.delete_keypair(keypair)
        self.assertIsNone(ret)

    def test_keypair(self):
        # create the keypair

        keypair = self.user_cloud.compute.create_keypair(
            name=self.keypair_name, type='ssh'
        )
        self.assertIsInstance(keypair, _keypair.Keypair)
        self.assertEqual(self.keypair_name, keypair.name)
        self.addCleanup(self._delete_keypair, keypair)

        # retrieve details of the keypair by ID

        keypair = self.user_cloud.compute.get_keypair(self.keypair_name)
        self.assertIsInstance(keypair, _keypair.Keypair)
        self.assertEqual(self.keypair_name, keypair.name)
        self.assertEqual(self.keypair_name, keypair.id)
        self.assertEqual('ssh', keypair.type)

        # retrieve details of the keypair by name

        keypair = self.user_cloud.compute.find_keypair(self.keypair_name)
        self.assertIsInstance(keypair, _keypair.Keypair)
        self.assertEqual(self.keypair_name, keypair.name)
        self.assertEqual(self.keypair_name, keypair.id)

        # list all keypairs

        keypairs = list(self.user_cloud.compute.keypairs())
        self.assertIsInstance(keypair, _keypair.Keypair)
        self.assertIn(self.keypair_name, {x.name for x in keypairs})


class TestKeypairAdmin(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()

        self.keypair_name = self.getUniqueString().split('.')[-1]
        self.user = self.operator_cloud.list_users()[0]

    def _delete_keypair(self, keypair):
        ret = self.user_cloud.compute.delete_keypair(keypair)
        self.assertIsNone(ret)

    def test_keypair(self):
        # create the keypair (for another user)
        keypair = self.operator_cloud.compute.create_keypair(
            name=self.keypair_name, user_id=self.user.id
        )
        self.assertIsInstance(keypair, _keypair.Keypair)
        self.assertEqual(self.keypair_name, keypair.name)
        self.addCleanup(self._delete_keypair, keypair)

        # retrieve details of the keypair by ID (for another user)

        keypair = self.operator_cloud.compute.get_keypair(
            self.keypair_name, user_id=self.user.id
        )
        self.assertEqual(self.keypair_name, keypair.name)
        self.assertEqual(self.keypair_name, keypair.id)
        self.assertEqual(self.user.id, keypair.user_id)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from openstack.network.v2 import qos_minimum_packet_rate_rule
from openstack.tests.unit import base


EXAMPLE = {
    'id': 'IDENTIFIER',
    'qos_policy_id': 'qos-policy-' + uuid.uuid4().hex,
    'min_kpps': 1500,
    'direction': 'any',
}


class TestQoSMinimumPacketRateRule(base.TestCase):
    def test_basic(self):
        sot = qos_minimum_packet_rate_rule.QoSMinimumPacketRateRule()
        self.assertEqual('minimum_packet_rate_rule', sot.resource_key)
        self.assertEqual('minimum_packet_rate_rules', sot.resources_key)
        self.assertEqual(
            '/qos/policies/%(qos_policy_id)s/minimum_packet_rate_rules',
            sot.base_path,
        )
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = qos_minimum_packet_rate_rule.QoSMinimumPacketRateRule(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['qos_policy_id'], sot.qos_policy_id)
        self.assertEqual(EXAMPLE['min_kpps'], sot.min_kpps)
        self.assertEqual(EXAMPLE['direction'], sot.direction)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# mypy: disable-error-code="method-assign"

from openstack.network.v2 import floating_ip
from openstack.network.v2 import network
from openstack.network.v2 import port
from openstack.network.v2 import router
from openstack.network.v2 import subnet
from openstack.tests.functional.network.v2 import common


class TestFloatingIP(common.TestTagNeutron):
    IPV4 = 4
    EXT_CIDR = "10.100.0.0/24"
    INT_CIDR = "10.101.0.0/24"
    EXT_NET_ID: str
    INT_NET_ID: str
    EXT_SUB_ID: str
    INT_SUB_ID: str
    ROT_ID: str
    PORT_ID: str
    FIP: floating_ip.FloatingIP
    DNS_DOMAIN = "example.org."
    DNS_NAME = "fip1"

    def setUp(self):
        super().setUp()
        if not self.user_cloud._has_neutron_extension("external-net"):
            self.skipTest(
                "Neutron external-net extension is required for this test"
            )
        self.TIMEOUT_SCALING_FACTOR = 1.5
        self.ROT_NAME = self.getUniqueString()
        self.INT_NET_NAME = self.getUniqueString()
        self.INT_SUB_NAME = self.getUniqueString()
        self.is_dns_supported = False

        # Find External Network
        for net in self.user_cloud.network.networks(is_router_external=True):
            self.EXT_NET_ID = net.id
        # Find subnet of the chosen external net
        for sub in self.user_cloud.network.subnets(network_id=self.EXT_NET_ID):
            self.EXT_SUB_ID = sub.id
        if not self.EXT_NET_ID and self.operator_cloud:
            # There is no existing external net, but operator
            # credentials available
            # WARNING: this external net is not dropped
            # Create External Network
            net = self._create_network(
                self.EXT_NET_NAME, **{"router:external": True}
            )
            self.EXT_NET_ID = net.id
            sub = self._create_subnet(
                self.EXT_SUB_NAME, self.EXT_NET_ID, self.EXT_CIDR
            )
            self.EXT_SUB_ID = sub.id

        # Create Internal Network
        net = self._create_network(self.INT_NET_NAME)
        self.INT_NET_ID = net.id
        sub = self._create_subnet(
            self.INT_SUB_NAME, self.INT_NET_ID, self.INT_CIDR
        )
        self.INT_SUB_ID = sub.id
        # Create Router
        sot = self.user_cloud.network.create_router(
            name=self.ROT_NAME,
            **{"external_gateway_info": {"network_id": self.EXT_NET_ID}},
        )
        assert isinstance(sot, router.Router)
        self.assertEqual(self.ROT_NAME, sot.name)
        self.ROT_ID = sot.id
        self.ROT = sot
        # Add Router's Interface to Internal Network
        sot = self.ROT.add_interface(
            self.user_cloud.network, subnet_id=self.INT_SUB_ID
        )
        self.assertEqual(sot["subnet_id"], self.INT_SUB_ID)
        # Create Port in Internal Network
        prt = self.user_cloud.network.create_port(network_id=self.INT_NET_ID)
        assert isinstance(prt, port.Port)
        self.PORT_ID = prt.id
        self.PORT = prt
        # Create Floating IP.
        fip_args = dict(
            floating_network_id=self.EXT_NET_ID,
        )
        if self.user_cloud._has_neutron_extension(
            "dns-integration"
        ) and self.user_cloud.has_service("dns"):
            self.is_dns_supported = True
            fip_args.update(
                dict(dns_domain=self.DNS_DOMAIN, dns_name=self.DNS_NAME)
            )
        fip = self.user_cloud.network.create_ip(**fip_args)
        assert isinstance(fip, floating_ip.FloatingIP)
        self.FIP = fip
        self.ID = self.FIP.id
        self.get_command = self.user_cloud.network.get_ip

    def tearDown(self):
        sot = self.user_cloud.network.delete_ip(
            self.FIP.id, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_port(
            self.PORT_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.ROT.remove_interface(
            self.user_cloud.network, subnet_id=self.INT_SUB_ID
        )
        self.assertEqual(sot["subnet_id"], self.INT_SUB_ID)
        sot = self.user_cloud.network.delete_router(
            self.ROT_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_subnet(
            self.INT_SUB_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_network(
            self.INT_NET_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def _create_network(self, name, **args):
        self.name = name
        net = self.user_cloud.network.create_network(name=name, **args)
        assert isinstance(net, network.Network)
        self.assertEqual(self.name, net.name)
        return net

    def _create_subnet(self, name, net_id, cidr):
        self.name = name
        self.net_id = net_id
        self.cidr = cidr
        sub = self.user_cloud.network.create_subnet(
            name=self.name,
            ip_version=self.IPV4,
            network_id=self.net_id,
            cidr=self.cidr,
        )
        assert isinstance(sub, subnet.Subnet)
        self.assertEqual(self.name, sub.name)
        return sub

    def test_find_by_id(self):
        sot = self.user_cloud.network.find_ip(self.FIP.id)
        self.assertEqual(self.FIP.id, sot.id)

    def test_find_by_ip_address(self):
        sot = self.user_cloud.network.find_ip(self.FIP.floating_ip_address)
        self.assertEqual(self.FIP.floating_ip_address, sot.floating_ip_address)
        self.assertEqual(self.FIP.floating_ip_address, sot.name)

    def test_find_available_ip(self):
        sot = self.user_cloud.network.find_available_ip()
        self.assertIsNotNone(sot.id)
        self.assertIsNone(sot.port_id)
        self.assertIsNone(sot.port_details)

    def test_get(self):
        sot = self.user_cloud.network.get_ip(self.FIP.id)
        self.assertEqual(self.EXT_NET_ID, sot.floating_network_id)
        self.assertEqual(self.FIP.id, sot.id)
        self.assertEqual(self.FIP.floating_ip_address, sot.floating_ip_address)
        self.assertEqual(self.FIP.fixed_ip_address, sot.fixed_ip_address)
        self.assertEqual(self.FIP.port_id, sot.port_id)
        self.assertEqual(self.FIP.port_details, sot.port_details)
        self.assertEqual(self.FIP.router_id, sot.router_id)
        if self.is_dns_supported:
            self.assertEqual(self.DNS_DOMAIN, sot.dns_domain)
            self.assertEqual(self.DNS_NAME, sot.dns_name)

    def test_list(self):
        ids = [o.id for o in self.user_cloud.network.ips()]
        self.assertIn(self.FIP.id, ids)

    def test_update(self):
        sot = self.user_cloud.network.update_ip(
            self.FIP.id, port_id=self.PORT_ID
        )
        self.assertEqual(self.PORT_ID, sot.port_id)
        self._assert_port_details(self.PORT, sot.port_details)
        self.assertEqual(self.FIP.id, sot.id)

    def _assert_port_details(self, port, port_details):
        self.assertEqual(port.name, port_details["name"])
        self.assertEqual(port.network_id, port_details["network_id"])
        self.assertEqual(port.mac_address, port_details["mac_address"])
        self.assertEqual(
            port.is_admin_state_up, port_details["admin_state_up"]
        )
        self.assertEqual(port.status, port_details["status"])
        self.assertEqual(port.device_id, port_details["device_id"])
        self.assertEqual(port.device_owner, port_details["device_owner"])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Extension(resource.Resource):
    resources_key = "extensions"
    base_path = "/extensions"

    # Capabilities
    allow_list = True

    #: Properties
    #: The alias for the extension.
    alias = resource.Body('alias', type=str)
    #: The extension description.
    description = resource.Body('description', type=str)
    #: Links pertaining to this extension.
    links = resource.Body('links', type=list)
    #: The name of this extension.
    name = resource.Body('name')
    #: The date and time when the resource was updated.
    #: The date and time stamp format is ISO 8601.
    updated_at = resource.Body('updated', type=str)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import tap_flow
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'name': 'my_tap_flow',
    'source_port': '1234',
    'tap_service_id': '4321',
    'id': IDENTIFIER,
    'project_id': '42',
}


class TestTapFlow(base.TestCase):
    def test_basic(self):
        sot = tap_flow.TapFlow()
        self.assertEqual('tap_flow', sot.resource_key)
        self.assertEqual('tap_flows', sot.resources_key)
        self.assertEqual('/taas/tap_flows', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = tap_flow.TapFlow(**EXAMPLE)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['source_port'], sot.source_port)
        self.assertEqual(EXAMPLE['tap_service_id'], sot.tap_service_id)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)

        self.assertDictEqual(
            {
                'limit': 'limit',
                'marker': 'marker',
                'name': 'name',
                'project_id': 'project_id',
                'sort_key': 'sort_key',
                'sort_dir': 'sort_dir',
            },
            sot._query_mapping._mapping,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ClusterTemplate(resource.Resource):
    resources_key = 'clustertemplates'
    base_path = '/clustertemplates'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_patch = True

    commit_method = 'PATCH'
    commit_jsonpatch = True

    #: The exposed port of COE API server.
    apiserver_port = resource.Body('apiserver_port', type=int)
    #: Display the attribute os_distro defined as appropriate metadata in image
    #: for the bay/cluster driver.
    cluster_distro = resource.Body('cluster_distro')
    #: Specify the Container Orchestration Engine to use. Supported COEs
    #: include kubernetes, swarm, mesos.
    coe = resource.Body('coe')
    #: The date and time when the resource was created.
    created_at = resource.Body('created_at')
    #: The name of a driver to manage the storage for the images and the
    #: container’s writable layer.
    docker_storage_driver = resource.Body('docker_storage_driver')
    #: The size in GB for the local storage on each server for the Docker
    #: daemon to cache the images and host the containers.
    docker_volume_size = resource.Body('docker_volume_size', type=int)
    #: The DNS nameserver for the servers and containers in the bay/cluster to
    #: use.
    dns_nameserver = resource.Body('dns_nameserver')
    #: The name or network ID of a Neutron network to provide connectivity to
    #: the external internet for the bay/cluster.
    external_network_id = resource.Body('external_network_id')
    #: The name or network ID of a Neutron network to provide connectivity to
    #: the internal network for the bay/cluster.
    fixed_network = resource.Body('fixed_network')
    #: Fixed subnet that are using to allocate network address for nodes in
    #: bay/cluster.
    fixed_subnet = resource.Body('fixed_subnet')
    #: The nova flavor ID or name for booting the node servers.
    flavor_id = resource.Body('flavor_id')
    #: The IP address for a proxy to use when direct http access
    #: from the servers to sites on the external internet is blocked.
    #: This may happen in certain countries or enterprises, and the
    #: proxy allows the servers and containers to access these sites.
    #: The format is a URL including a port number. The default is
    #: None.
    http_proxy = resource.Body('http_proxy')
    #: The IP address for a proxy to use when direct https access from the
    #: servers to sites on the external internet is blocked.
    https_proxy = resource.Body('https_proxy')
    #: The name or UUID of the base image in Glance to boot the servers for the
    #: bay/cluster.
    image_id = resource.Body('image_id')
    #: The URL pointing to users’s own private insecure docker
    #: registry to deploy and run docker containers.
    insecure_registry = resource.Body('insecure_registry')
    #: Whether enable or not using the floating IP of cloud provider.
    is_floating_ip_enabled = resource.Body('floating_ip_enabled')
    #: Indicates whether the ClusterTemplate is hidden or not.
    is_hidden = resource.Body('hidden', type=bool)
    #: this option can be set to false to create a bay/cluster without the load
    #: balancer.
    is_master_lb_enabled = resource.Body('master_lb_enabled', type=bool)
    #: Specifying this parameter will disable TLS so that users can access the
    #: COE endpoints without a certificate.
    is_tls_disabled = resource.Body('tls_disabled', type=bool)
    #: Setting this flag makes the baymodel/cluster template public and
    #: accessible by other users.
    is_public = resource.Body('public', type=bool)
    #: This option provides an alternative registry based on the Registry V2
    is_registry_enabled = resource.Body('registry_enabled', type=bool)
    #: The name of the SSH keypair to configure in the bay/cluster servers for
    #: ssh access.
    keypair_id = resource.Body('keypair_id')
    #: Arbitrary labels. The accepted keys and valid values are defined in the
    #: bay/cluster drivers. They are used as a way to pass additional
    #: parameters that are specific to a bay/cluster driver.
    labels = resource.Body('labels', type=dict)
    #: The flavor of the master node for this baymodel/cluster template.
    master_flavor_id = resource.Body('master_flavor_id')
    #: The name of a network driver for providing the networks for the
    #: containers.
    network_driver = resource.Body('network_driver')
    #: When a proxy server is used, some sites should not go through the proxy
    #: and should be accessed normally.
    no_proxy = resource.Body('no_proxy')
    #: The servers in the bay/cluster can be vm or baremetal.
    server_type = resource.Body('server_type')
    #: The date and time when the resource was updated.
    updated_at = resource.Body('updated_at')
    #: The UUID of the cluster template.
    uuid = resource.Body('uuid', alternate_id=True)
    #: The name of a volume driver for managing the persistent storage for the
    #: containers.
    volume_driver = resource.Body('volume_driver')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.network.v2 import l3_conntrack_helper as _l3_conntrack_helper
from openstack.network.v2 import router
from openstack.tests.functional import base


class TestL3ConntrackHelper(base.BaseFunctionalTest):
    PROTOCOL = "udp"
    HELPER = "tftp"
    PORT = 69

    ROT_ID = None

    def setUp(self):
        super().setUp()

        if not self.user_cloud.network.find_extension("l3-conntrack-helper"):
            self.skipTest("L3 conntrack helper extension disabled")

        self.ROT_NAME = self.getUniqueString()
        # Create Router
        sot = self.user_cloud.network.create_router(name=self.ROT_NAME)
        self.assertIsInstance(sot, router.Router)
        self.assertEqual(self.ROT_NAME, sot.name)
        self.ROT_ID = sot.id
        self.ROT = sot

        # Create conntrack helper
        ct_helper = self.user_cloud.network.create_conntrack_helper(
            router=self.ROT,
            protocol=self.PROTOCOL,
            helper=self.HELPER,
            port=self.PORT,
        )
        self.assertIsInstance(ct_helper, _l3_conntrack_helper.ConntrackHelper)
        self.CT_HELPER = ct_helper

    def tearDown(self):
        sot = self.user_cloud.network.delete_router(
            self.ROT_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_get(self):
        sot = self.user_cloud.network.get_conntrack_helper(
            self.CT_HELPER, self.ROT_ID
        )
        self.assertEqual(self.PROTOCOL, sot.protocol)
        self.assertEqual(self.HELPER, sot.helper)
        self.assertEqual(self.PORT, sot.port)

    def test_list(self):
        helper_ids = [
            o.id
            for o in self.user_cloud.network.conntrack_helpers(self.ROT_ID)
        ]
        self.assertIn(self.CT_HELPER.id, helper_ids)

    def test_update(self):
        NEW_PORT = 90
        sot = self.user_cloud.network.update_conntrack_helper(
            self.CT_HELPER.id, self.ROT_ID, port=NEW_PORT
        )
        self.assertEqual(NEW_PORT, sot.port)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import access_rule
from openstack.tests.unit import base

EXAMPLE = {
    "links": {
        "self": "https://example.com/identity/v3/access_rules"
        "/07d719df00f349ef8de77d542edf010c"
    },
    "path": "/v2.1/servers/{server_id}/ips",
    "method": "GET",
    "service": "compute",
}


class TestAccessRule(base.TestCase):
    def test_basic(self):
        sot = access_rule.AccessRule()
        self.assertEqual('access_rule', sot.resource_key)
        self.assertEqual('access_rules', sot.resources_key)
        self.assertEqual('/users/%(user_id)s/access_rules', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = access_rule.AccessRule(**EXAMPLE)
        self.assertEqual(EXAMPLE['path'], sot.path)
        self.assertEqual(EXAMPLE['method'], sot.method)
        self.assertEqual(EXAMPLE['service'], sot.service)
        self.assertEqual(EXAMPLE['links'], sot.links)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.dns.v2 import _base
from openstack import exceptions
from openstack import resource
from openstack import utils


class Zone(_base.Resource):
    """DNS ZONE Resource"""

    resources_key = 'zones'
    base_path = '/zones'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    commit_method = "PATCH"

    _query_mapping = resource.QueryParameters(
        'name',
        'type',
        'email',
        'status',
        'description',
        'ttl',
        'limit',
        'marker',
    )

    #: Properties
    #: current action in progress on the resource
    action = resource.Body('action')
    #: Attributes
    #: Key:Value pairs of information about this zone, and the pool the user
    #: would like to place the zone in. This information can be used by the
    #: scheduler to place zones on the correct pool.
    attributes = resource.Body('attributes', type=dict)
    #: Timestamp when the zone was created
    created_at = resource.Body('created_at')
    #: Zone description
    #: *Type: str*
    description = resource.Body('description')
    #: The administrator email of this zone
    #: *Type: str*
    email = resource.Body('email')
    #: Links contains a `self` pertaining to this zone or a `next` pertaining
    #: to next page
    links = resource.Body('links', type=dict)
    #: The master list for slaver server to fetch DNS
    masters = resource.Body('masters', type=list)
    #: Zone name
    name = resource.Body('name')
    #: The pool which manages the zone, assigned by system
    pool_id = resource.Body('pool_id')
    #: The project id which the zone belongs to
    project_id = resource.Body('project_id')
    #: Serial number in the SOA record set in the zone,
    #: which identifies the change on the primary DNS server
    #: *Type: int*
    serial = resource.Body('serial', type=int)
    #: Zone status
    #: Valid values include `PENDING_CREATE`, `ACTIVE`,
    #: `PENDING_DELETE`, `ERROR`
    status = resource.Body('status')
    #: SOA TTL time, unit is seconds, default 300, TTL range 300-2147483647
    #: *Type: int*
    ttl = resource.Body('ttl', type=int)
    #: Zone type,
    #: Valid values include `PRIMARY`, `SECONDARY`
    #: *Type: str*
    type = resource.Body('type')
    #: Timestamp when the zone was last updated
    updated_at = resource.Body('updated_at')
    #: Whether the zone is shared with other projects
    #: *Type: bool*
    is_shared = resource.Body('shared')

    # Headers for DELETE requests
    #: If true, delete any existing zone shares along with the zone
    delete_shares = resource.Header('x-designate-delete-shares', type=bool)

    def _action(self, session, action, body):
        """Preform actions given the message body."""
        url = utils.urljoin(self.base_path, self.id, 'tasks', action)
        response = session.post(url, json=body)
        exceptions.raise_from_response(response)
        return response

    def abandon(self, session):
        self._action(session, 'abandon', None)

    def xfr(self, session):
        self._action(session, 'xfr', None)
